GCKI = require('common')('GCKI')

GCKI.dprint("Entered control.lua")
require("util")

if script.active_mods["gvv"] then
  require("__gvv__.gvv")()
  log("Enabled support for gvv!")
end

------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
--                                 LOCAL DEFINITIONS                              --
------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
local area = {}
-- If the distance between owner and vehicle is greater than this value, we will
-- output messages for the owner.
local message_owner_distance = 30

--------------------------------------------------------------------------------------------
-- Get version of the base game. As long as we can't assume that everybody uses a version --
-- of Factorio that allows filtering script_raised_X events (>= 0.18.27), we need to make --
-- sure that the installed game can fall back.
--------------------------------------------------------------------------------------------
local function allow_script_raised_filters()
    local f_name = "allow_script_raised_filters"
    GCKI.dprint("Entered function " .. f_name)

    local base = game and game.active_mods["base"]
    global.filter_script_raised = false

      -- Base mod is active
    if base then
        local version = util.split(base, '.')
GCKI.dprint("Factorio version: " .. serpent.line(version))
        if tonumber(version[1]) == 0 then
            -- This test will only be necessary before Factorio 1.0
            if ( ((tonumber(version[2]) or 0) == 18) and (tonumber(version[3]) or 0) >= 27) then
                global.filter_script_raised = true
            end
        -- Filtering of script-raised events will be safe to use once Factorio 1.0 is out!
        else
          global.filter_script_raised = true
        end

    end
GCKI.dprint("Allow filtering of script raised events: " .. serpent.line(global.filter_script_raised))
    GCKI.dprint("End of function " .. f_name)
end

------------------------------------------------------------------------------------
-- Function to measure distance between vehicle's current position and position it
-- has been summoned to.
local function get_distance(cur_pos, new_pos)
  local x, y = cur_pos.x-new_pos.x, cur_pos.y-new_pos.y
  return math.sqrt(x*x+y*y)
end

------------------------------------------------------------------------------------
--                     CHOOSE THE OCCUPANT WITH WEAPONS CONTROL                   --
-- If there are two players inside a vehicle, Autodrive hands over gun control to --
-- the driver, without checking which player actually is in charge of the weapons.--
-- vehicle: vehicle.entity or vehicle.unit_number
------------------------------------------------------------------------------------
local function set_shooter(vehicle) -- vehicle: entity
    local f_name = "set_shooter"
    --~ GCKI.dprint("Entered function " .. f_name .." (" .. tostring(vehicle and vehicle.name) .. ")")
    GCKI.dprint("Entered function " .. f_name .." (" .. GCKI.argprint(vehicle) .. ")")

    --~ if not (vehicle and vehicle.valid and vehicle.type == "car") then
        --~ GCKI.error("Invalid vehicle " .. GCKI.argprint(vehicle) .. "!", f_name)
    --~ end

    if (vehicle and vehicle.valid and vehicle.type == "car") then
        local occupants = GCKI.get_occupants(vehicle)

        if occupants then
            -- If the vehicle has a passenger and this passenger is no dummy inserted by other
            -- mods, he will be in charge of the weapons
                if occupants["passenger"] and occupants["passenger"].is_player() then
                    vehicle.driver_is_gunner = false
                else
                    vehicle.driver_is_gunner = true
                end
        --~ -- If nobody is in the vehicle, make sure the Autodrive passenger can shoot!
        else
            vehicle.driver_is_gunner = false
        end
    end
    --~ GCKI.dprint("End of function " .. f_name .." (" .. tostring(vehicle and vehicle.name) .. ")")
    GCKI.dprint("End of function " .. f_name .." (" .. GCKI.argprint(vehicle) .. ")")
end


------------------------------------------------------------------------------------
--       CHECK WHETHER VEHICLE IS CONTROLLED BY MODS LIKE AUTODRIVE OR AAI        --
--                             (Returns array of dummies or nil)                             --
------------------------------------------------------------------------------------
local function is_dummy_inside(vehicle) -- vehicle: entity
    local f_name = "is_dummy_inside"
    GCKI.dprint("Entered function " .. f_name .. " (" .. GCKI.argprint(vehicle) .. ")")

    --~ if not (vehicle and vehicle.valid and vehicle.type == "car") then
        --~ GCKI.error("Invalid vehicle " .. GCKI.argprint(vehicle) .. "!", f_name)
    --~ end

    -- Initialize return value
    local dummy = {}

    if (vehicle and vehicle.valid and vehicle.type == "car") then

        -- Only check occupied vehicles for dummies
        local occupants = GCKI.get_occupants(vehicle)
        if not occupants then
            dummy = nil
        --~ end

        --~ dummy = {}
        else
            for role, occupant in pairs(occupants) do
                if occupant and occupant.valid then
                    -- Player entity without a name (dummy)
    GCKI.dprint("Occupant name: " .. tostring(occupant.name)  .. "Player: " .. tostring(occupant.is_player()))
                    if occupant.is_player() and not occupant.name then
                        dummy[role] = occupant
    GCKI.dprint(role .. " is a dummy player. dummy[" .. role .. "]: " .. tostring(dummy.role))
                    -- Player entity with name (real player)
                    elseif occupant.is_player() then
    GCKI.dprint(role .. " is a real player. dummy[" .. role .. "]: " .. tostring(dummy.role))

                    -- Occupant  is a character connected to a real player.
                    elseif occupant.player then
    GCKI.dprint(role .. " is a character connected to a player. dummy[" .. role .. "]: " .. tostring(dummy.role))
                    -- Put occupant on dummy list if it's neither a player nor connected to a player.
                    else
    GCKI.dprint(role .. " is a dummy character. dummy[" .. role .. "]: " .. tostring(dummy.role))
                        dummy[role] = occupant
                    end
                end
            end
        end
    end
    GCKI.dprint("End of function " .. f_name .. " (" .. GCKI.argprint(vehicle) .. "). RETURN: " ..              serpent.line(dummy))
    return (dummy and table_size(dummy) > 0) and dummy or nil
    --~ return dummy
end

------------------------------------------------------------------------------------
--                        CHECK WHETHER VEHICLE HAS AN OWNER                      --
--                     (Returns the player who owns it, or nil)                   --
------------------------------------------------------------------------------------
local function is_owned_vehicle(vehicle) -- vehicle: entity or entity.unit_nuber
    local f_name = "is_owned_vehicle"
    GCKI.dprint("Entered function " .. f_name .. " (" .. GCKI.argprint(vehicle) .. ")")

    local v_id = (type(vehicle) == "number" and vehicle) or
                 (vehicle and vehicle.valid and vehicle.type == "car" and vehicle.unit_number) or
                 --~ GCKI.error("Invalid vehicle " .. GCKI.argprint(vehicle) .. "!", f_name)
                 nil
    local v = v_id and global.GCKI_vehicles[v_id]

    local owner = v and v.owner and game.players[v.owner]

    GCKI.dprint("End of function " .. f_name .. " (" .. GCKI.argprint(vehicle) .. "). RETURN: " ..              serpent.line(owner))
    return owner
end

------------------------------------------------------------------------------------
--                       CHECK WHETHER PLAYER OWNS A VEHICLE                      --
--          (Returns unit_number of vehicle owned by the player, or nil)          --
------------------------------------------------------------------------------------
local function owned_by_player(player) -- player.index or player entity
    local f_name = "owned_by_player"
    GCKI.dprint("Entered function " .. f_name .. " (" .. GCKI.argprint(player) .. ")")


    --~ local index = (type(player) == "number" and game.players[player] and player) or
                  --~ (player and player.valid and player.is_player() and player.index) or
                  --~ GCKI.error("Invalid player " .. GCKI.argprint(player), f_name)
    local index = (type(player) == "number" and player) or
                  (player and player.valid and player.is_player() and player.index) or
                  nil
GCKI.show(index, "index")

    local vehicle = nil

    if index then
        for v, data in pairs(global.GCKI_vehicles) do
            if data.owner and data.owner == index then
                vehicle = v
                break
            end
        end
    end
    GCKI.dprint("End of function " .. f_name .. " (" .. GCKI.argprint(player) .. "). RETURN: " ..               serpent.line(vehicle))
    return vehicle
end

------------------------------------------------------------------------------------
--                          CHECK WHETHER VEHICLE IS LOCKED                       --
--                    (Returns the player who locked it, or nil)                  --
------------------------------------------------------------------------------------
local function is_locked_vehicle(vehicle)
    local f_name = "is_locked_vehicle"
    GCKI.dprint("Entered function " .. f_name .. " (" .. GCKI.argprint(vehicle) .. ")")

    local v_id = (type(vehicle) == "number" and vehicle) or
                 (vehicle and vehicle.valid and vehicle.type == "car" and vehicle.unit_number) or
                 --~ GCKI.error("Invalid vehicle " .. GCKI.argprint(vehicle) .. "!", f_name)
                 nil

    --~ if not (vehicle and vehicle.valid) then
        --~ GCKI.error("Invalid vehicle " .. GCKI.argprint(vehicle), f_name)
    --~ end

    local v = v_id and global.GCKI_vehicles[v_id]
    local locker = v and v.locker and game.players[v.locker]

    GCKI.show(locker and locker.name, "locker")
--~ GCKI.dprint("global table: " .. serpent.block(global))
    GCKI.dprint("End of function " .. f_name .. " (" .. GCKI.argprint(vehicle) .. "). RETURN: " ..             serpent.line(locker))
    return locker
end

------------------------------------------------------------------------------------
--                          GET VEHICLES LOCKED BY PLAYER                         --
--     (Returns table of unit_numbers of vehicle owned by the player, or nil)     --
------------------------------------------------------------------------------------
local function locked_by_player(player) -- player.index or player entity
    local f_name = "locked_by_player"
    GCKI.dprint("Entered function " .. f_name .. " (" .. GCKI.argprint(player) .. ")")

    local index = (type(player) == "number" and game.players[player] and player) or
                  (player and player.valid and player.is_player() and player.index) or
                  --~ GCKI.error("Invalid player " .. GCKI.argprint(player), f_name)
                  nil
    local locked_vehicles = {}

    --~ for player, vehicle2 in pairs(global.vehicles) do
        --~ if vehicle == vehicle2 then
            --~ owner = game.players[player]
            --~ break
        --~ end
    --~ end
    if index then
        for v, data in pairs(global.GCKI_vehicles) do
            if data.locker and data.locker == index then
                table.insert(locked_vehicles, v)
            end
        end
    end
    local ret = (table_size(locked_vehicles) > 0) and locked_vehicles or nil
    GCKI.dprint("End of function " .. f_name .. " (" .. GCKI.argprint(player) .. "). RETURN: " ..               serpent.line(ret))
    return ret
end

------------------------------------------------------------------------------------
--                        SET/UNSET LOCKED-ICON ON VEHICLE                        --
------------------------------------------------------------------------------------
local function locked_icon(vehicle)
    local f_name = "locked_icon"
    GCKI.dprint("Entered function " .. f_name .. " (" .. GCKI.argprint(vehicle) .. ")")

    local v_id = (type(vehicle) == "number" and vehicle) or
                 (vehicle and vehicle.valid and vehicle.type == "car" and vehicle.unit_number) or
                 --~ GCKI.error("Invalid vehicle " .. GCKI.argprint(vehicle) .. "!", f_name)
                 nil
    local v = v_id and global.GCKI_vehicles[v_id]

    if v then
        -- Set icon
        if v.locker and not v.lock_icon then
            v.lock_icon = rendering.draw_sprite{
                sprite = "GCKI_locked-icon",
                target = v.entity,
                target_offset = {0, -0.5},
                surface = v.entity.surface,
                render_layer = "entity-info-icon-above",
                only_in_alt_mode = true,
            }
        -- Remove icon
        elseif v.lock_icon and not v.locker then
            rendering.destroy(v.lock_icon)
            v.lock_icon = nil
        end
    end
    GCKI.dprint("End of function " .. f_name .. " (" .. GCKI.argprint(vehicle) .. ")")
end


------------------------------------------------------------------------------------
--                 CHECK WHETHER VEHICLE MUST BE REMOVED FROM LIST                --
--     We only need to store vehicles that have either locker, owner or both!     --
--                             (Returns boolean value)                            --
------------------------------------------------------------------------------------
local function must_remove_vehicle(vehicle) -- vehicle.unit_number or vehicle entity
    local f_name = "must_remove_vehicle"
    GCKI.dprint("Entered function " .. f_name .. " (" .. GCKI.argprint(vehicle) .. ")")

    local v_id = (type(vehicle) == "number" and vehicle) or
                 (vehicle and vehicle.valid and vehicle.type == "car" and vehicle.unit_number) or
                 --~ GCKI.error("Invalid vehicle " .. GCKI.argprint(vehicle) .. "!", f_name)
                 nil
    local v = v_id and global.GCKI_vehicles[v_id]

    local owner, locker

    if v and v.entity and v.entity.valid then
        owner = is_owned_vehicle(v.entity)
        locker = is_locked_vehicle(v.entity)
    end

    local ret = (not owner and not locker)
    GCKI.dprint("End of function " .. f_name .. " (" .. GCKI.argprint(vehicle) .. "). RETURN: " ..             serpent.line(ret))
    return ret
end

------------------------------------------------------------------------------------
--                      CLEAN UP TABLES WHEN VEHICLE IS REMOVED                   --
------------------------------------------------------------------------------------
local function remove_vehicle(vehicle)
    local f_name = "remove_vehicle"
    GCKI.dprint("Entered function " .. f_name .. " (" .. tostring(vehicle) .. ")")

    local v = (type(vehicle) == "number" and vehicle) or
              (vehicle and vehicle.valid and vehicle.type == "car" and vehicle.unit_number) or
              --~ GCKI.error("Invalid vehicle " .. GCKI.argprint(vehicle), f_name)
              nil

    if v then
        global.GCKI_vehicles[v] = nil
    end

    GCKI.dprint("End of function " .. f_name .. " (" .. tostring(vehicle) .. ")")
end


------------------------------------------------------------------------------------
--                        MARK VEHICLE AS LOCKED BY PLAYER                        --
--    Each player may only own one vehicle at any time, but may lock his vehicle  --
--    before exchanging it, and may thus, in effect, own more than one vehicle.   --
------------------------------------------------------------------------------------
local function add_vehicle_locker(vehicle, player) -- vehicle.unit_number or vehicle entity
                                                   -- player.index or player entity
    local f_name = "add_vehicle_locker"
    GCKI.dprint("Entered function " .. f_name .. " (" .. GCKI.argprint(vehicle) ..
                ", " .. GCKI.argprint(player) .. ")")

    local v_id = (type(vehicle) == "number" and vehicle) or
                 (vehicle and vehicle.valid and vehicle.type == "car" and vehicle.unit_number) or
                 GCKI.error("Invalid vehicle " .. GCKI.argprint(vehicle) .. "!", f_name)
    local v = global.GCKI_vehicles[v_id]

    local player_id = (type(player) == "number" and game.players[player] and player) or
                      (player and player.valid and player.is_player() and player.index) or
                      GCKI.error("Invalid player " .. GCKI.argprint(player), f_name)

    local owner = v and is_owned_vehicle(v.entity)

    -- Vehicle is already in list -- add "locker"
    if v then
        v.locker = player_id
    -- Add vehicle to list
    else
        global.GCKI_vehicles[v_id] = {
            ["entity"] = vehicle,
            -- This isn't needed: if there was an owner, the vehicle would already be in the list!
            ["owner"] = owner and owner.index,
            ["locker"] = player_id
        }
    end

    -- Notify "Autodrive" that vehicle has been locked
    if remote.interfaces["autodrive"] then
        remote.call("autodrive", "set_GCKI_locker", v_id, player_id)
    end


    GCKI.dprint("End of function " .. f_name .. " (" .. GCKI.argprint(vehicle) ..
                ", " .. GCKI.argprint(player) .. ")")
end

------------------------------------------------------------------------------------
--                            MARK VEHICLE AS UNLOCKED                            --
------------------------------------------------------------------------------------
local function remove_vehicle_locker(vehicle, player)
    local f_name = "remove_vehicle_locker"
    GCKI.dprint("Entered function " .. f_name .. " (" .. GCKI.argprint(vehicle) ..
                ", " .. GCKI.argprint(player) .. ")")

    local v_id = (type(vehicle) == "number" and vehicle) or
                 (vehicle and vehicle.valid and vehicle.type == "car" and vehicle.unit_number) or
                 --~ GCKI.error("Invalid vehicle " .. GCKI.argprint(vehicle) .. "!", f_name)
                 nil
    local v = v_id and global.GCKI_vehicles[v_id]

    local player_id = (type(player) == "number" and game.players[player] and player) or
                      (player and player.valid and player.is_player() and player.index) or
                      --~ GCKI.error("Invalid player " .. GCKI.argprint(player), f_name)
                      nil

    --~ if v and v.locker == player_id then
        --~ v.locker = nil
    --~ end
    if v and v.locker and v.locker == player_id then
        --~ if v.locker then
            -- Notify "Autodrive" that vehicle has no locker
            if remote.interfaces["autodrive"] then
                remote.call("autodrive", "set_GCKI_locker", v_id, nil)
            end
            --~ -- Notify "Vehicle Wagons" that vehicle has been unlocked
            --~ if remote.interfaces["VehicleWagon2"] then
                --~ remote.call("VehicleWagon2", "release_locked_by_player", v.locker)
            --~ end
            -- Remove locker from vehicle data
            --~ if player_id and v.locker and v.locker == player_id then
            v.locker = nil
            --~ end
        --~ end

        -- Remove unlocked vehicle from list if nobody owns it
        if must_remove_vehicle(v.entity) then
            remove_vehicle(v)
        end
    end

    GCKI.dprint("End of function " .. f_name .. " (" .. GCKI.argprint(vehicle) ..
                ", " .. GCKI.argprint(player) .. ")")
end

------------------------------------------------------------------------------------
--                        MARK VEHICLE AS OWNED BY PLAYER                         --
--                Each player may only own one vehicle at any time!               --
------------------------------------------------------------------------------------
local function add_vehicle_owner(vehicle, player) -- vehicle: vehicle entity
                                                  -- player: player.index or player entity
    local f_name = "add_vehicle_owner"
    GCKI.dprint("Entered function " .. f_name .. " (" .. GCKI.argprint(vehicle) ..
                ", " .. GCKI.argprint(player) .. ")")

    local v_id = (vehicle and vehicle.valid and vehicle.type == "car" and vehicle.unit_number) or
                 GCKI.error("Invalid vehicle " .. GCKI.argprint(vehicle) .. "!", f_name)
    local v = global.GCKI_vehicles[v_id]

    local player_id = (type(player) == "number" and game.players[player] and player) or
                      (player and player.valid and player.is_player() and player.index) or
                      GCKI.error("Invalid player " .. GCKI.argprint(player), f_name)

    local locker = v and is_locked_vehicle(v.entity)

    -- Player can only own a locked vehicle if he also locked it!
    if locker and locker.index ~= player_id then
        GCKI.error("New owner (" .. player_id .. ") is not the player who locked vehicle (" ..
                    locker.index .. ")", f_name)
    end

    -- Vehicle is already in list -- add "owner"
    if v then
        v.owner = player_id
    -- Add vehicle to list
    else
        global.GCKI_vehicles[v_id] = {
            ["entity"] = vehicle,
            ["owner"] = player_id,
            ["locker"] = locker and locker.index
        }
    end

    -- Notify "Autodrive" that vehicle now has an owner
    if remote.interfaces["autodrive"] then
        remote.call("autodrive", "set_GCKI_owner", v_id, player_id)
    end

    -- Notify "Vehicle Wagons" that vehicle has no owner
    if remote.interfaces["VehicleWagon2"] then
GCKI.dprint("Calling VehicleWagon2 because owner of vehicle " .. tostring(v_id) .. " ("..player_id..") has claimed another vehicle.")
        remote.call("VehicleWagon2", "release_owned_by_player", player_id)
    end
    GCKI.dprint("End of function " .. f_name .. " (" .. GCKI.argprint(vehicle) ..
                ", " .. GCKI.argprint(player) .. ")")
end

------------------------------------------------------------------------------------
--                            MARK VEHICLE AS UNOWNED                             --
------------------------------------------------------------------------------------
local function remove_vehicle_owner(vehicle, player)
    local f_name = "remove_vehicle_owner"
    GCKI.dprint("Entered function " .. f_name .. " (" .. GCKI.argprint(vehicle) ..
                ", " .. GCKI.argprint(player) .. ")")

    local v_id = (type(vehicle) == "number" and vehicle) or
                 (vehicle and vehicle.valid and vehicle.type == "car" and vehicle.unit_number) or
                 --~ GCKI.error("Invalid vehicle " .. GCKI.argprint(vehicle) .. "!", f_name)
                 nil
    local v = v_id and global.GCKI_vehicles[v_id]

    local player_id = (type(player) == "number" and game.players[player] and player) or
                      (player and player.valid and player.is_player() and player.index) or
                      --~ GCKI.error("Invalid player " .. GCKI.argprint(player), f_name)
                      nil

    --~ if v and v.owner == player_id then
        --~ v.owner = nil
    --~ end
    if v then
        -- Notify "Autodrive" that vehicle has no owner
        if remote.interfaces["autodrive"] then
GCKI.dprint("Calling Autodrive because owner of vehicle " .. tostring(v_id) .. " has claimed another vehicle.")
            remote.call("autodrive", "set_GCKI_owner", v_id, nil)
        end

        -- Notify "Vehicle Wagons" that vehicle has no owner
        if remote.interfaces["VehicleWagon2"] then
GCKI.dprint("Calling VehicleWagon2 because owner of vehicle " .. tostring(v_id) .. " has claimed another vehicle.")
            remote.call("VehicleWagon2", "release_owned_by_player", v.owner)
        end

        -- Remove owner from vehicle data
        --~ if player_id and v.locker and v.locker == player_id then
        v.owner = nil
        --~ end

        -- Remove vehicle from list if it's unlocked
        if must_remove_vehicle(v_id) then
            remove_vehicle(v_id)
        end
    end

    GCKI.dprint("End of function " .. f_name .. " (" .. GCKI.argprint(vehicle) ..
                ", " .. GCKI.argprint(player) .. ")")
end

--~ -- This is probably useless now that we have the replace_vehicle function:
--~ ------------------------------------------------------------------------------------
--~ --      CLEAN UP LIST OF VEHICLES LOCKED BY PLAYER BY REMOVING INVALID VEHICLES   --
--~ ------------------------------------------------------------------------------------
--~ local function remove_invalid_locked_vehicles(player) -- player.index or player entity
    --~ local f_name = "remove_invalid_locked_vehicles()"
    --~ GCKI.dprint("Entered function remove_invalid_locked_vehicles(" .. tostring(player and player.name) .. ")")

    --~ if not (vehicle and vehicle.valid) then
        --~ GCKI.error("Invalid player " .. tostring(player), f_name)
    --~ end

    --~ local index = player.index
    --~ if global.locked_vehicles[index] and (table_size(global.locked_vehicles[index]) > 0) then
        --~ for pos,vehicle in pairs(global.locked_vehicles[index]) do
            --~ if not vehicle.valid then
                --~ global.locked_vehicles[index][pos] = nil
            --~ end
        --~ end
    --~ end
    --~ GCKI.dprint("End of function remove_invalid_locked_vehicles(" .. tostring(player and player.name) .. ")")
--~ end


------------------------------------------------------------------------------------
--                             LOCK OR UNLOCK A VEHICLE                           --
------------------------------------------------------------------------------------
-- vehicle.unit_number or vehicle entity, player.index or player
local function lock_or_unlock(vehicle, player)
    local f_name = "lock_or_unlock"
    GCKI.dprint("Entered function " .. f_name .. " (" .. GCKI.argprint(vehicle) .. ")")

    local v_id = (type(vehicle) == "number" and vehicle) or
                 (vehicle and vehicle.valid and vehicle.type == "car" and vehicle.unit_number) or
                 (vehicle and vehicle.entity and vehicle.entity.valid and vehicle.entity.type == "car" and vehicle.entity.unit_number) or
                --~ GCKI.error("Invalid vehicle " .. GCKI.argprint(vehicle) .. "!", f_name)
                 nil
    local player = (type(player) == "number" and game.players[player] and player) or
                (player and player.is_player() and player.index) or
                GCKI.error("Invalid player " .. GCKI.argprint(player) .. "!")

    local v_data = v_id and global.GCKI_vehicles[v_id]
    local v = v_data and v_data.entity
GCKI.show(v_id, "v_id")
GCKI.show(v and v.valid, "v.valid")
GCKI.show(v and v.valid and v.name, "v.name")
GCKI.show(player, "player")
    if v and v.valid then
GCKI.dprint("vehicle active: " .. serpent.line(v.active))
GCKI.dprint("not vehicle active: " .. serpent.line(not v.active))
GCKI.dprint("vehicle operable: " .. serpent.line(v.operable))
GCKI.dprint("vehicle minable: " .. serpent.line(v.minable))
      -- Unlock vehicle on account of locker
      if v_data.locker and v_data.locker == player then
GCKI.dprint("Unlocking " .. GCKI.argprint(v))
        v.active = true
        v.operable = true
        -- If setting from Unminable Vehicles is enabled, no vehicle may be mined
        v.minable = true
        if global.UnminableVehicles then
          v.minable = false
        end
        --~ v_data.locker = nil
        remove_vehicle_locker(v_id, player)
      -- Don't unlock a vehicle that has been locked by somebody else!
      elseif v_data.locker and v_data.locker ~= player then
GCKI.dprint("Can't unlock " .. GCKI.argprint(v) .. " because it has been locked by " .. game.get_player(v_data.locker).name)
      -- Lock vehicle on account of player
      else
GCKI.dprint("Locking " .. GCKI.argprint(v))
        v.active = false
        v.operable = false
        v.minable = false
        --~ v_data.locker = player
        add_vehicle_locker(v_id, player)
      end

GCKI.dprint("Toggled active, minable, and operable!")
GCKI.dprint("vehicle active: " .. serpent.line(v.active))
GCKI.dprint("vehicle operable: " .. serpent.line(v.operable))
GCKI.dprint("vehicle minable: " .. serpent.line(v.minable))
      locked_icon(v_id)
    end

GCKI.show(is_locked_vehicle(vehicle), "is_locked_vehicle(vehicle)")
    GCKI.dprint("End of function " .. f_name .. " (" .. GCKI.argprint(vehicle) .. ")")
end

------------------------------------------------------------------------------------
--                          GET NAMES OF VEHICLE OCCUPANTS                        --
--       (Returns table with {"driver" = LuaPlayer, "passenger" = LuaPlayer})     --
------------------------------------------------------------------------------------
local function get_occupant_names(vehicle)      -- vehicle: entity
    local f_name = "get_occupant_names"
    GCKI.dprint("Entered function " .. f_name .. " (" .. GCKI.argprint(vehicle) .. ")")


    --~ if not (vehicle and vehicle.valid and vehicle.type == "car") then
        --~ GCKI.error("Invalid vehicle " .. GCKI.argprint(vehicle), f_name)
    --~ end
    if not (vehicle and vehicle.valid and vehicle.type == "car") then
        vehicle = nil
    end

    local occupants = vehicle and GCKI.get_occupants(vehicle)

    if  occupants then
        if occupants["driver"] and occupants["passenger"] then
            occupants = {"GCKI-words.and", occupants["driver"].name, occupants["passenger"].name}
        else
            occupants = (occupants["driver"] and occupants["driver"].name) or
                        (occupants["passenger"] and occupants["passenger"].name)
        end
    end

    GCKI.dprint("End of function " .. f_name .. " (" .. GCKI.argprint(vehicle) .. "). RETURN: " ..              serpent.line(occupants))
    return occupants
end


------------------------------------------------------------------------------------
--                CHANGE VEHICLE COLOR TO COLOR OF OWNER OR LOCKER                --
------------------------------------------------------------------------------------
local function color_vehicle(vehicle, ...) -- v: entity
    local f_name = "color_vehicle"
    GCKI.dprint("Entered function " .. f_name .. " (" .. GCKI.argprint(vehicle) .. ")")
    --~ GCKI.dprint("Entered function " .. f_name .. " (" .. GCKI.argprint(vehicle) .. ")")

    --~ local v_id = (type(v) == "table" and v.valid and v.type == "car" and v.unit_number) or
                 --~ GCKI.error("Invalid vehicle " .. GCKI.argprint(v), f_name)
    local v = vehicle and vehicle.valid and (vehicle.type == "car") and vehicle

    if v then
        local owner = is_owned_vehicle(v)
        local locker = is_locked_vehicle(v)

        local autodrive_owner = (... and type(...) == "number") and game.players[...]
        if not autodrive_owner then
            autodrive_owner = remote.interfaces["autodrive"] and
                              remote.call("autodrive", "get_vehicle_owner", v)
            autodrive_owner = autodrive_owner and game.players[autodrive_owner]
        end

        -- Set color to owner
        if owner then
GCKI.dprint("Changing color of " .. tostring(v.name) .. " (" .. v.unit_number ..
            ") to color of owner (" .. tostring(owner.index) .. ")")
            v.color = owner.color
        -- Set color to locker
        elseif locker then
GCKI.dprint("Changing color of " .. tostring(v.name) .. " (" .. v.unit_number ..
            ") to color of owner (" .. tostring(locker.index) .. ")")
            v.color = locker.color
        -- Set color to autodrive_owner
        elseif autodrive_owner then
GCKI.dprint("Changing color of " .. tostring(v.name) .. " (" .. v.unit_number ..
            ") to color of autodrive_owner (" .. tostring(autodrive_owner.index) .. ")")
            v.color = autodrive_owner.color
        -- Remove color
        else
GCKI.dprint("Changing color of " .. tostring(v.name) .. " (" .. v.unit_number .. ") to default color")
            v.color = GCKI.colors["no_color"]
        end
    end
    --~ GCKI.dprint("End of function " .. f_name .. " (" .. GCKI.argprint(vehicle) .. ")")
    GCKI.dprint("End of function " .. f_name .. " (" .. GCKI.argprint(vehicle) .. ")")
end



------------------------------------------------------------------------------------
--             REPLACE VEHICLE IN LISTS OF CLAIMED AND LOCKED VEHICLES            --
------------------------------------------------------------------------------------
local function replace_vehicle(event)   -- table
    local f_name = "replace_vehicle"
    GCKI.dprint("Entered function " .. f_name .. " (" .. serpent.line(event) .. ")")

    local v_new = event.new_entity
    local v_new_id = v_new.unit_number

    local v_old_id = event.old_entity_unit_number

    local locker = global.GCKI_vehicles[v_old_id] and global.GCKI_vehicles[v_old_id].locker
    local owner = global.GCKI_vehicles[v_old_id] and global.GCKI_vehicles[v_old_id].owner

    --~ add_vehicle(v_new, { ["locker"] = locker, ["owner"] = owner })

    -- Remove dummy characters from vehicle
    local dummy = is_dummy_inside(v_new)
    if dummy then
        for role, character in pairs(dummy) do
            GCKI.dprint({"", "Removed dummy ", role, " ", tostring(character and character.name),
                            " from ", GCKI.loc_name(v_new), "!"})
            character.destroy()
        end
    end

    -- Add new vehicle to list
    global.GCKI_vehicles[v_new_id] = {
        ["entity"] = v_new,
        ["locker"] = locker,
        ["owner"] = owner
    }

    -- Recolor new vehicle
    color_vehicle(v_new)

    -- Lock new vehicle if it has a locker
    if locker then
        v_new.active = false
        v_new.minable = false
        v_new.operable = false

        locked_icon(v_new)
    end

    -- Remove old vehicle from list
    remove_vehicle(v_old_id)
--~ GCKI.dprint("removed old vehicle " .. tostring(v_old_id))
--~ GCKI.dprint("global table: " .. serpent.block(global))

    GCKI.dprint("End of function " .. f_name .. " (" .. serpent.line(event) .. ")")
end

------------------------------------------------------------------------------------
--                                VEHICLE RESTORED                                --
--    This is needed for "Vehicle Wagons 2", when it unloads a vehicle that was   --
--                   previously locked or owned by some player.                   --
------------------------------------------------------------------------------------
local function restore_vehicle(v, v_data)
    local f_name = "restore_vehicle"
    GCKI.dprint("Entered function " .. f_name .. " (" .. GCKI.argprint(v) .. ", " ..
                GCKI.argprint(v_data) .. ")")
    local vehicle = (v and v.valid and v.type == "car" and v) or
                     GCKI.error("Invalid v_data " .. GCKI.argprint(v_data.entity), f_name)
    --~ local owner = v_data and v_data.owner and
                  --~ ((type(v_data.owner) == "number" and v_data.owner) or
                   --~ (v_data.owner.valid and v_data.owner.is_player() and v_data.owner.index)) or
                  --~ nil
    local owner = v_data and v_data.owner and
                  ((type(v_data.owner) == "number" and v_data.owner) or
                   (v_data.owner.valid and v_data.owner.is_player() and v_data.owner.index)) or
                  nil
    local locker = v_data and v_data.locker and
                   ((type(v_data.locker) == "number" and v_data.locker) or
                    (v_data.locker.valid and v_data.locker.is_player() and v_data.locker.index)) or
                   nil
--~ GCKI.dprint("TEST -- querying owned_by_player: " .. tostring(remote.call("GCKI", "owned_by_player", owner or 1)))
--~ GCKI.dprint("Restoring vehicle: " .. tostring(vehicle.name) .. " (" .. tostring(vehicle.unit_number) .. ")")
--~ GCKI.dprint("owner: " .. serpent.block(owner))
--~ GCKI.dprint("locker: " .. serpent.block(locker))
    -- Player may only own one vehicle. If he owns a new one, override "owner" from stored data!
    local new_owned = owner and owned_by_player(owner)
    if new_owned and new_owned ~= vehicle.unit_number then
--~ GCKI.dprint("Player stored as owner of " .. vehicle.unit_number .. " in vehicle data ("
--~ .. owner .. ") has claimed another vehicle (" .. new_owned .. ")")
        owner = nil
    end

    -- Add v_data to list (Either locker or owner must exist; owner may not have claimed a new vehicle.)
    --~ if  locker or (owner and not owned_by_player(owner)) then
    if  locker or owner then
--~ GCKI.dprint("Adding  vehicle " .. tostring(vehicle.name) .. " (" .. tostring(vehicle.unit_number) .. ") to GCKI-list")
        global.GCKI_vehicles[vehicle.unit_number] = {
            entity = vehicle,
            locker = locker,
            owner = owner
        }
        GCKI.dprint("Added vehicle on account of Vehicle Wagons 2: " ..
                    serpent.block(global.GCKI_vehicles[vehicle.unit_number]))

        -- Lock vehicle if it has a locker
        if locker then
            vehicle.active = false
            vehicle.minable = false
            vehicle.operable = false

            locked_icon(vehicle)
            GCKI.dprint("Locked v_data " .. vehicle.name ..
                        "(" .. vehicle.unit_number .. ") for " .. game.players[locker].name)
        end

        -- Inform owner that he unloaded his vehicle
        if owner then
            GCKI.output(game.players[owner], {"GCKI-messages.vehicle_unloaded_owner", GCKI.loc_name(vehicle)})
        end

    end
    -- Reset vehicle color
    color_vehicle(vehicle)
    GCKI.dprint("End of function " .. f_name .. " (" .. GCKI.argprint(v_data) .. ")")
end


------------------------------------------------------------------------------------
--       REMOVE DUMMIES (FIX FOR AAI PROGRAMMABLE VEHICLES AND SIMILAR MODS)      --
-- Mods may put a dummy driver or passenger into a vehicle if it's unoccupied. It --
-- should be destroyed again if it's not needed anymore -- but if it isn't        --
-- when the vehicle dies, the dummy character is left on the surface, so we have  --
-- to remove it.                                                                  --
------------------------------------------------------------------------------------
local function remove_dummy(vehicle) -- vehicle: entity
    local f_name = "remove_dummy"
    GCKI.dprint("Entered function " .. f_name .. " (" .. GCKI.argprint(vehicle) .. ")")

    local v = (vehicle and vehicle.valid and vehicle.type == "car" and vehicle) or
              --~ GCKI.error("Invalid vehicle " .. GCKI.argprint(vehicle), f_name)
              nil

    local dummy = (v and is_dummy_inside(v)) or {}

    for role, character in pairs(dummy) do
        GCKI.dprint({"", "Removed dummy ", role, " ", tostring(character and character.name),
                        " from ", GCKI.loc_name(v), " (on_entity_died)!"})
        character.destroy()
    end

    GCKI.dprint("End of function " .. f_name .. " (" .. GCKI.argprint(vehicle) .. ")")
end

------------------------------------------------------------------------------------
--               RENDER MESSAGES TO A PARTICULAR PLAYER ON THE GROUND             --
------------------------------------------------------------------------------------
local function text_on_ground(args)
    local f_name = "text_on_ground"
    GCKI.dprint("Entered function " .. f_name .. " (" .. serpent.line(args) .. ")")

    for _,varname in pairs{"text", "players", "vehicle"} do
        if args[varname] == nil then
            GCKI.error("Missing " .. varname .. "!", f_name)
        elseif args[varname].valid ~= nil and args[varname].valid == false then
            GCKI.error("Invalid Lua object for " .. varname .. "!", f_name)
        end
    end

    local msg = args.text
    local player = args.players
    local vehicle = args.vehicle
    local time_to_live = global.player_settings[player.index]["render_time"]
    -- Render message on ground
    local id = rendering.draw_text{
                            text            = msg,
                            surface         = vehicle.surface,
                            target          = vehicle.position,
                            color           = global.player_settings[player.index].render_color,
                            players         = {player},
                            time_to_live    = time_to_live,
                            alignment       = "center"
                        }
    -- If recipient of the message is online, add the message to his list of messages
    if player.connected then
GCKI.dprint("Text to ground: " .. player.name .. " is connected.")
        table.insert(global.render_msgs[player.index], {ID = id, TTL = time_to_live})
    -- If recipient of the message is offline, add the message to his list and set its
    -- time_to_live to 0 (message won't expire)
    else
GCKI.dprint("Text to ground: " .. player.name .. " is offline.")
        table.insert(global.render_msgs[player.index], {ID = id, TTL = time_to_live})
        rendering.set_time_to_live(id, 0)
    end
    GCKI.dprint("End of function " .. f_name .. " (" .. serpent.line(args) .. ")")
end

------------------------------------------------------------------------------------
-- REMOVE EXPIRED MESSAGES RENDERED ON THE GROUND FROM A PARTICULAR PLAYER'S LIST --
------------------------------------------------------------------------------------
local function remove_expired_text_on_ground(player) -- player: player.index or entity
    local f_name = "remove_expired_text_on_ground"
    GCKI.dprint("Entered function " .. f_name .. " (" .. GCKI.argprint(player) .. ")")

    local player_id = (type(player) == "number" and game.players[player] and player) or
                      (player and player.valid and player.is_player() and player.index) or
                      --~ GCKI.error("Invalid player " .. GCKI.argprint(player), f_name)
                      nil

    --~ if not player and player.valid then
        --~ GCKI.error("Missing or invalid player " .. GCKI.argprint(player) .. "!", f_name)
    --~ end

    --~ local index = player.index
    if player_id then
--~ GCKI.show(global.render_msgs[index], "global.render_msgs[" .. index .. "]")
        for x, msg in ipairs (global.render_msgs[player_id]) do
    --~ GCKI.show(msg, "msg")
            if not rendering.is_valid(msg.ID) then
    --~ GCKI.dprint("Removing message with id " .. tostring(msg.ID))
                global.render_msgs[player_id][x] = nil
            end
        end
    end
--~ GCKI.show(global.render_msgs[index], "global.render_msgs[" .. index .. "] (after removal of invalid entries)")
    GCKI.dprint("End of function " .. f_name .. " (" .. GCKI.argprint(player) .. ")")
end

------------------------------------------------------------------------------------
--            CREATE AN AREA AROUND CLICK-POSITION TO SEARCH FOR VEHICLES         --
------------------------------------------------------------------------------------
local function area_create(position,dx,dy)
    if not dy then dy = dx end
    return {{position.x - dx, position.y - dy},{position.x + dx, position.y + dy}}
end

------------------------------------------------------------------------------------
--                                  INITIALIZE GAME                               --
------------------------------------------------------------------------------------
local function init()
    GCKI.dprint("Entered function init()")

    global.ticks_since_action = global.ticks_since_action or 0

    -- Restore or create list of vehicles currently owned or locked by somebody
    -- (sorted by vehicle.unit_number)
    global.GCKI_vehicles = global.GCKI_vehicles or {}

    -- Restore or create list with render messages for connected players
    -- (sorted by player index)
    global.render_msgs = global.render_msgs or {}

    -- Restoring per-user runtime settings would overwrite any changes to these
    -- settings the user made just before loading/joining the game. Yet, the settings
    -- must be stored in the global table because all player's instances of Factorio
    -- must know about the other players -- otherwise, desync errors *will* happen!
    global.player_settings = global.player_settings or {}
    -- Restore startup settings
    GCKI.get_startup_settings()
    -- Restore global runtime settings
    GCKI.get_global_settings()

--~ game.check_prototype_translations()
    -- Compatibility with old Factorio versions (<0.18.27)
    if global.filter_script_raised == nil then
        global.filter_script_raised = allow_script_raised_filters()
    end

    -- Compatibility with other mods
    --~ -- We want to know if AAI Programmable vehicles is installed, so we can remove
    --~ -- its dummy characters from the surface if the mod is removed from the game.
    --~ global.AAI = global.AAI or game.active_mods["aai-programmable-vehicles"] and true or false
    --~ GCKI.dprint("Added flag for AAI to global table.")

    -- Is "Uncle Shred's unminable vehicles" active?
    global.UnminableVehicles = script.active_mods["UnminableVehicles"] and
                               settings.global["unminable_vehicles_make_unminable"].value
GCKI.dprint("global: " .. serpent.line(global))

    GCKI.dprint("End of function init()")
end

------------------------------------------------------------------------------------
--                            INITIALIZE PLAYER SETTINGS                          --
------------------------------------------------------------------------------------
local function init_player(player)
    local f_name = "init_player"
    GCKI.dprint("Entered function " .. f_name .. " (" .. GCKI.argprint(player) .. ")")

    local player_id = (type(player) == "number" and game.players[player] and player) or
                      (player and player.valid and player.is_player() and player.index) or
                      GCKI.error("Invalid player " .. GCKI.argprint(player), f_name)

    -- Restore or create list with render messages for connected players
    -- (Contains array of messages: {ID = id, TTL = time_to_live})
    global.render_msgs[player_id] = global.render_msgs[player_id] or {}

    -- Restore user settings
    GCKI.get_user_settings(player_id)

    -- Check whether we can filter script-raised events
    if global.filter_script_raised == nil then
        global.filter_script_raised = allow_script_raised_filters()
    end

    GCKI.dprint("End of function " .. f_name .. " (" .. GCKI.argprint(player) .. ")")
end

------------------------------------------------------------------------------------
--                 CLEAN UP TABLES WHEN PLAYER IS REMOVED FROM GAME               --
------------------------------------------------------------------------------------
local function remove_player(player) -- player: player.index or player entity
    local f_name = "remove_player"
    GCKI.dprint("Entered function " .. f_name .. "(" .. GCKI.argprint(player) .. ")")

    local player_id = (type(player) == "number" and game.players[player] and player) or
                      (player and player.valid and player.is_player() and player.index) or
                      --~ GCKI.error("Invalid player " .. GCKI.argprint(player), f_name)
                      nil
    --~ local p = game

    if player_id then
        -- Get vehicle owned by player
        local v_owned = owned_by_player(player)
        -- Get vehicles locked by player
        local v_locked = locked_by_player(player)

        -- Remove list of player runtime settings
        global.player_settings[player_id] = nil
        GCKI.dprint("Removed list of user runtime settings of " .. player.name)

        -- Remove player as owner of vehicle in list
        if v_owned then
            local vehicle = global.GCKI_vehicles[v_owned].entity
            GCKI.dprint(player.name .. " is no longer the owner of " ..
                        vehicle.name .. " (" .. v_owned ..  ")!")
            global.GCKI_vehicles[v_owned] = nil
            color_vehicle(vehicle)
            -- Remove vehicle from list if it has neither owner nor locker
            if must_remove_vehicle(v_owned) then
                remove_vehicle(v_owned)
            end
        end

        -- Remove player as locker of vehicles in list and unlock vehicles
        if v_locked then
            local vehicle
            for _, v in ipairs(v_locked) do
                vehicle = global.GCKI_vehicles[v]
                --~ lock_or_unlock(vehicle.entity)
                --~ vehicle.locker = nil
                lock_or_unlock(vehicle.entity, player)
                --~ vehicle.locker = nil
                color_vehicle(vehicle.entity)

                if must_remove_vehicle(v) then
                    remove_vehicle(v)
                end
                GCKI.dprint({"", "Unlocked ", GCKI.loc_name(vehicle), " locked by ", player.name})
            end
        end
        --  Remove messages rendered on the ground rendered for the player
        for _, msg in ipairs(global.render_msgs[player_id] or {}) do
            rendering.destroy(msg.ID)
            GCKI.dprint("Destroyed message with ID " .. tostring(id) ..
                            " rendered for player " .. player.name .. ".")
        end
        -- Remove list of messages on the ground rendered for the player
        global.render_msgs[player_id] = nil
        GCKI.dprint("Removed list of messages rendered on the ground for " .. player.name)
    end
    GCKI.dprint("End of function " .. f_name .. "(" .. GCKI.argprint(player) .. ")")
end

------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
--                                   EVENT SCRIPTS                                --
------------------------------------------------------------------------------------
------------------------------------------------------------------------------------


------------------------------------------------------------------------------------
--                        TIMER (ONLY ONE ACTION PER SECOND)                      --
------------------------------------------------------------------------------------
script.on_event(defines.events.on_tick, function(event)
    global.ticks_since_action = global.ticks_since_action + 1
end)


------------------------------------------------------------------------------------
--                                INITIALIZE NEW GAME                             --
------------------------------------------------------------------------------------
script.on_init(function(event)
    GCKI.dprint("Entered event script on_init")

    init()

    GCKI.dprint("End of event script on_init")
end)

------------------------------------------------------------------------------------
--                                GENERATE NEW PLAYER                             --
------------------------------------------------------------------------------------
script.on_event(defines.events.on_player_created, function(event)
    GCKI.dprint("Entered event script on_player_created for player " .. game.players[event.player_index].name)
    local player = game.players[event.player_index]

    init_player(player)

    GCKI.dprint("End of event script on_player_created for player " .. game.players[event.player_index].name)
end)

------------------------------------------------------------------------------------
--                            PLAYER (RE-)JOINED THE GAME                         --
------------------------------------------------------------------------------------
script.on_event(defines.events.on_player_joined_game, function(event)
    local index = event.player_index
    local player = game.players[index]
    GCKI.dprint("Entered event script on_player_joined_game for player " .. player.name)

    init_player(player)

    -- Set remaining time for messages on ground (keep msgs in table, they will be
    -- removed if they've become invalid next time the player disconnects)
    GCKI.dprint("Rendering messages to ".. player.name)
    for _, msg in pairs (global.render_msgs[index]) do
        -- Set time_to_live for existing message from 0 to remaining time
        if rendering.is_valid(msg.ID) then
            rendering.set_time_to_live(msg.ID, msg.TTL)
        end
    end
    GCKI.dprint("End of event script on_player_joined_game for player " .. player.name)
end)

------------------------------------------------------------------------------------
--                         PLAYER LEFT THE GAME TEMPORARILY                       --
------------------------------------------------------------------------------------
script.on_event(defines.events.on_pre_player_left_game, function(event)
    local index = event.player_index
    local player = game.players[index]
    GCKI.dprint("Entered event script on_pre_player_left_game for player " .. player.name)

    -- Clean up list of rendered messages for player
    remove_expired_text_on_ground(player)
    -- Replace TTL of each message stored in list of rendered messages for player
    -- with remaining TTL
    for _, msg in pairs (global.render_msgs[index]) do
        if rendering.is_valid(msg.ID) then
            msg.TTL = rendering.get_time_to_live(msg.ID)
            -- Set time_to_live for existing message to 0 (msg won't expire)
            rendering.set_time_to_live(msg.ID, 0)
        end
    end
--~ GCKI.dprint("global table: " .. serpent.block(global))
    GCKI.dprint("End of event script on_pre_player_left_game for player " .. player.name)
end)

------------------------------------------------------------------------------------
--                             PERMANENTLY DELETE PLAYER                          --
------------------------------------------------------------------------------------
script.on_event(defines.events.on_pre_player_removed, function(event)
    local player = game.players[event.player_index]
    GCKI.dprint("Entered event script on_pre_player_removed for player " .. player.name)

    remove_player(player)
    GCKI.output(player, "You have been deleted from the game. Bye, bye!")
--~ GCKI.dprint("global table: " .. serpent.block(global))

    GCKI.dprint("End of event script on_pre_player_removed for player " .. player.name)
end)

------------------------------------------------------------------------------------
--                               CONFIGURATION CHANGED                            --
------------------------------------------------------------------------------------
script.on_configuration_changed(function(event)
    GCKI.dprint("Entered event script on_configuration_changed")

    init()
    for _, player in pairs(game.players) do
        init_player(player)
    end

    ------------------------------------------------------------------------------------
    -- Enable this when "AAI Programmable Vehicles" has been updated and names
    -- dummy characters properly. Otherwise, we might destroy anonymous characters
    -- created by other mods.

    --~ -- Check whether AAI was in the game before but has been removed.
    --~ GCKI.dprint("on_configuration_changed event: " .. serpent.block(event))
    --~ local AAI = event.mod_changes["aai-programmable-vehicles"]

    --~ -- Search for character entities that are not connected to any player (AAI dummies)
    --~ if AAI and AAI.old_version and not AAI.new_version then
        --~ local characters = {}
        --~ for _, s in pairs(game.surfaces) do
            --~ characters = s.find_entities_filtered{ type = "character" }
            --~ for c, character in pairs(characters) do
                --~ if not character.player then
                    --~ GCKI.dprint("Removing character " .. character.name .. "/" ..character.unit_number ..
                                --~ " (not connected to player).")
                    --~ character.destroy()
                --~ end
            --~ end
        --~ end
    --~ end

    ------------------------------------------------------------------------------------
    -- Compatibility with "Uncle Shred's unminable vehicles" -- make sure that even unlocked
    -- vehicles are unminable if that mod is active.

    global.UnminableVehicles = script.active_mods["UnminableVehicles"] and
               settings.global["unminable_vehicles_make_unminable"] and
               settings.global["unminable_vehicles_make_unminable"].value

    for _, v in pairs(global.GCKI_vehicles or {}) do
GCKI.dprint("Checking vehicle " .. _ .. " (" .. tostring(v.entity and v.entity.valid and v.entity.name or "not valid") .. "):" .. serpent.block(v) )
      -- Locked vehicles are still unminable!
      if v.entity and v.entity.valid then
GCKI.show(v.locker, "v.locker")
GCKI.show(v.entity.active, "v.entity.active")
        v.entity.active = v.locker and false or true
GCKI.show(v.entity.active, "Set v.entity.active to")
        v.entity.operable = v.entity.active
        -- If "Unminable Vehicles" is active, vehicles will always be unminable!
        v.entity.minable = v.entity.active
        if global.UnminableVehicles then
          v.entity.minable = false
        end

        -- Remove icon from unlocked vehicles
        if v.lock_icon and not v.locker then
          rendering.destroy(v.lock_icon)
          v.lock_icon = nil
GCKI.dprint("Removed lock_icon!")
        end
      -- Remove invalid vehicles from list
      else
        if v.lock_icon then
          rendering.destroy(v.lock_icon)
GCKI.dprint("Removed lock_icon!")
        end
        global.GCKI_vehicles[_] = nil
GCKI.dprint("Removed invalid vehicle " .. tostring(_))
      end
    end

    GCKI.dprint("End of event script on_configuration_changed")
end)

------------------------------------------------------------------------------------
--                              RUNTIME SETTING CHANGED                           --
------------------------------------------------------------------------------------
script.on_event(defines.events.on_runtime_mod_setting_changed, function(event)
    GCKI.dprint("Entered event script on_runtime_mod_setting_changed")

        local player = event.player_index
        local setting = event.setting
GCKI.dprint("Changed settings: " .. serpent.block(event))

    -- GCKI settings have been changed
    if string.find(setting, "^GCKI-") then
        -- Change runtime-global settings
        if event.setting_type == "runtime-global" then
            GCKI.get_global_settings()
            GCKI.dprint("Changed global runtime settings!")
        -- Change runtime-per-user settings
        else
            GCKI.get_new_user_setting(player, setting)
            GCKI.dprint("Changed runtime user setting \"" .. tostring(setting) .. "\" for player " ..
                            tostring(game.get_player(player).name) .. ".")
        end
    --~ -- Setting of "Uncle Shred's unminable vehicles" has been changed
    elseif event.setting == "unminable_vehicles_make_unminable" then

        global.UnminableVehicles = settings.global["unminable_vehicles_make_unminable"].value
        if not global.UnminableVehicles then
            for _, v in pairs(global.GCKI_vehicles) do
GCKI.dprint("Checking vehicle " .. _ .. " (" .. v.entity.name .. ") unminable:" .. serpent.block(v) )
            -- Only locked vehicles are unminable
                if v.locker then
GCKI.dprint("Making vehicle unminable because it has been locked")
                    v.entity.minable = false
                else
GCKI.dprint("Making vehicle minable again")
                    v.entity.minable = true
                end
            end
        end
    end

    GCKI.dprint("End of event script on_runtime_mod_setting_changed")
end)

------------------------------------------------------------------------------------
--                             A PLAYER STOLE A VEHICLE                           --
------------------------------------------------------------------------------------
script.on_event(defines.events.on_pre_player_mined_item, function(event)
    GCKI.dprint("Entered event script on_pre_player_mined_item for player " ..
                    game.players[event.player_index].name)

    local vehicle = event.entity
    -- Even with filtered events, this is still necessary to prevent acting on events raised by scripts!
    if not vehicle.valid or vehicle.type ~= "car" then
        GCKI.dprint("Leaving event script on_pre_player_mined_item early (nothing to do for " .. vehicle.name .. ")")
        return
    end

    local player = game.players[event.player_index]
    local owner = vehicle.valid and is_owned_vehicle(vehicle)

    -- Only check for owner is needed, as locked vehicles can't be mined
    if owner then
        -- If the owner himself mined the vehicle, no message on the ground is needed
        if owner == player then
            GCKI.output(owner, {"GCKI-messages.vehicle_no_more"})
        -- Somebody else stole the vehicle, so inform the owner in any possible way!
        else
            -- Normal output
            GCKI.output(owner, {"GCKI-messages.miner_player", player.name, GCKI.loc_name(vehicle)})
            -- Does the owner want to have messages on the ground?
            if global.player_settings[owner.index].render_text then
                text_on_ground({
                    text = {"GCKI-render-messages.miner_player", GCKI.loc_name(vehicle), player.name},
                    players = owner,
                    vehicle = vehicle
                })
            end
        end
        -- Remove vehicle from tables
        remove_vehicle(vehicle)
    end

    GCKI.dprint("End of event script on_pre_player_mined_item for player " .. player.name)
-- Event filtering doesn't work for events raised by scripts, so we need to include filtering code
-- anyway. Let's just remove event filtering for now -- then the mod can also be used with the current stable
-- version of Factorio (0.17.74 as of 2019-11-19).
end, { {filter = "type", type = "car"} })
--~ end)

------------------------------------------------------------------------------------
--                               BOTS STOLE A VEHICLE                             --
------------------------------------------------------------------------------------
script.on_event(defines.events.on_robot_pre_mined, function(event)
    GCKI.dprint("Entered event script on_robot_pre_mined for entity " .. tostring(event.entity.name))

    local vehicle = event.entity
    -- Even with filtered events, this is still necessary to prevent acting on events raised by scripts!
    if (not vehicle.valid) or vehicle.type ~= "car" then
        GCKI.dprint("Leaving event script on_robot_pre_mined early (nothing to do for " .. vehicle.name .. ")")
        return
    end

    local owner = vehicle.valid and is_owned_vehicle(vehicle)

    -- Only check for owner is needed, as locked vehicles can't be mined
    if owner then
        GCKI.output(owner, {"GCKI-messages.miner_bot", GCKI.loc_name(vehicle)})
        -- Does the owner want to have messages on the ground?
        if global.player_settings[owner.index].render_text then
            -- Normal output
            text_on_ground({
                text = {"GCKI-render-messages.miner_bot", GCKI.loc_name(vehicle)},
                players = owner,
                vehicle = vehicle
            })
        end
        -- Remove vehicle from tables
        remove_vehicle(vehicle)
    end

    GCKI.dprint("End of event script on_robot_pre_mined for entity " .. tostring(event.entity.name))
-- Event filtering doesn't work for events raised by scripts, so we need to include filtering code
-- anyway. Let's just remove event filtering for now -- then the mod can also be used with the current stable
-- version of Factorio (0.17.74 as of 2019-11-19).
end, { {filter = "type", type = "car"} })
--~ end)

-- Removed for now!
--~ ------------------------------------------------------------------------------------
--~ --              VEHICLE WAS DESTROYED BY PLACING LANDFILL/WATERFILL               --
--~ ------------------------------------------------------------------------------------
--~ script.on_event({defines.events.on_player_built_tile,
                 --~ defines.events.on_robot_built_tile}, function(event)
    --~ GCKI.dprint("Entered event script on_X_built_tile for " .. tostring(event.tile.name))
--~ GCKI.dprint("Event data: " .. serpent.block(event))
    --~ local vehicle = event.entity
    --~ -- Even with filtered events, this is still necessary to prevent acting on events raised by scripts!
    --~ if (not vehicle.valid) or vehicle.type ~= "car" then
        --~ GCKI.dprint("Leaving event script on_X_mined_entity early (nothing to do for " .. vehicle.name .. ")")
        --~ return
    --~ end

    --~ local owner = vehicle.valid and is_owned_vehicle(vehicle)
    --~ local locker = vehicle.valid and is_locked_vehicle(vehicle)

    --~ local miner = event.player_index and game.players[event.player_index]
    --~ local bot = event.robot and true or false
    --~ local msg = {}

    --~ -- Entities can be mined if land-/waterfill is placed over them -- even if they are locked!
    --~ if owner then
        --~ msg = bot and {"GCKI-messages.miner_bot", GCKI.loc_name(vehicle)} or
                      --~ {"GCKI-messages.miner_player", tostring(miner and miner.name), GCKI.loc_name(vehicle)}
    --~ elseif locker then
        --~ msg = bot and {"GCKI-messages.miner_bot_locked", GCKI.loc_name(vehicle)} or
                      --~ {"GCKI-messages.miner_player_locked", tostring(miner and miner.name), GCKI.loc_name(vehicle)}
    --~ end

    --~ local player = owner or locker
    --~ if player then
        --~ GCKI.output(player, msg)
        --~ -- Does the owner want to have messages on the ground?
        --~ if global.player_settings[player.index].render_text then
            --~ -- Normal output
            --~ text_on_ground({
                --~ text = msg,
                --~ players = player,
                --~ vehicle = vehicle
            --~ })
        --~ end
        --~ -- Remove vehicle from tables
        --~ remove_vehicle(vehicle)
    --~ end

    --~ GCKI.dprint("End of event script on_X_built_tile for " .. tostring(event.tile.name))
-- Event filtering doesn't work for events raised by scripts, so we need to include filtering code
-- anyway. Let's just remove event filtering for now -- then the mod can also be used with the current stable
--~ -- version of Factorio (0.17.74 as of 2019-11-19).
--~ --end, { {filter = "type", type = "car"} })
--~ end)

-----------------------------------------------------------------------------------
--                               VEHICLE WAS DESTROYED                            --
------------------------------------------------------------------------------------
local function on_entity_died(event)
--~ script.on_event({defines.events.on_entity_died}, function(event)
    GCKI.dprint("Entered event script on_entity_died for entity " .. tostring(event.entity.name))
    local vehicle = event.entity
    -- Even with filtered events, this is still necessary to prevent acting on events raised by scripts!
    if (not vehicle.valid) or vehicle.type ~= "car" then
        GCKI.dprint("Leaving event script on_entity_died early (nothing to do for " .. vehicle.name .. ")")
        return
    end

    local owner = is_owned_vehicle(vehicle)
    local locker = is_locked_vehicle(vehicle)
    local players = {}

    local msg_text = {}
    local render_text = {}

    ------------------------------------------------------------------------------------
    --               FIX FOR AAI PROGRAMMABLE VEHICLES AND SIMILAR MODS               --
    ------------------------------------------------------------------------------------
    remove_dummy(vehicle)

    ------------------------------------------------------------------------------------
    -- Build message for owner
    if owner then
        msg_text = {"GCKI-messages.dead", GCKI.loc_name(vehicle)}
        render_text = {"GCKI-render-messages.dead", GCKI.loc_name(vehicle)}
        players = owner
    -- Build message for locker (will not be rendered if this vehicle is owned by someone)
    elseif locker then
        msg_text = {"GCKI-messages.dead_locker", GCKI.loc_name(vehicle), vehicle.surface.name,
                    vehicle.position.x, vehicle.position.y}
        render_text = {"GCKI-render-messages.dead_locker", GCKI.loc_name(vehicle)}
        players = locker
    end

    -- Print to owner/locker and render message for that player on the ground
    if owner or locker then
        GCKI.output(players, msg_text)
        text_on_ground({text = render_text, players = players, vehicle = vehicle})
    end

    -- Remove vehicle from tables
    remove_vehicle(vehicle)

    GCKI.dprint("End of event script on_entity_died for entity " .. tostring(event.entity.name))
-- Event filtering doesn't work for events raised by scripts, so we need to include filtering code
-- anyway. Let's just remove event filtering for now -- then the mod can also be used with the current stable
-- version of Factorio (0.17.74 as of 2019-11-19).
--~ end, { {filter = "type", type = "car"} })
--~ end)
end
script.on_event(defines.events.on_entity_died, on_entity_died, { {filter = "type", type = "car"} })
script.on_event(defines.events.script_raised_destroy, on_entity_died, global.filter_script_raised and
                    { {filter = "type", type = "car"} } or nil)

--~ ------------------------------------------------------------------------------------
--~ --                         VEHICLE WAS DESTROYED BY SCRIPT                        --
--~ ------------------------------------------------------------------------------------
--~ script.on_event(defines.events.script_raised_destroy, function(event)
    --~ GCKI.dprint("Entered event script script_raised_destroy for entity " .. serpent.block(event.entity.name))
--~ GCKI.dprint("event data: " .. serpent.block(event))

    --~ on_entity_died(event)

    --~ GCKI.dprint("End of event script script_raised_destroy for entity " .. tostring(event.entity.name))
--~ end)

------------------------------------------------------------------------------------
--                          VEHICLE WAS PLACED BY SCRIPT                          --
------------------------------------------------------------------------------------
script.on_event(defines.events.script_raised_built, function(event)
    GCKI.dprint("Entered event script script_raised_built for entity " .. serpent.block(event.entity.name))
GCKI.dprint("event data: " .. serpent.block(event))
    -- Need to keep the check in for the 0.17 branch!
    if event.entity and event.entity.valid and event.entity.type == "car" then

        -- Don't check for event.GCKI_data: If it isn't present, we may still need to
        -- recolor the vehicle, which is also done in restore_vehicle()!
        GCKI.dprint("Restore vehicle on account of mod " ..      serpent.line(event.mod_name))
        restore_vehicle(event.entity, event.GCKI_data)

    end
    GCKI.dprint("End of event script script_raised_built for entity " .. tostring(event.entity.name))
end, global.filter_script_raised and { {filter = "type", type = "car"} } or nil)

------------------------------------------------------------------------------------
--                           ENTITY BUILT (KEY WAS USED)                          --
------------------------------------------------------------------------------------
script.on_event(defines.events.on_built_entity, function(event)
    GCKI.dprint("Entered event script on_built_entity for player " .. game.players[event.player_index].name)
GCKI.dprint("event data: " .. serpent.block(event))
GCKI.dprint("created entity: " .. serpent.block(event.created_entity and event.created_entity.name))
GCKI.dprint("used item: " .. serpent.block(event.item and event.item.name))
--~ GCKI.dprint("Event data: " .. serpent.block(event))
    local index = event.player_index
    local player = game.players[index]
    local entity = event.created_entity or event.entity

    ------------------------------------------------------------------------------------
    -- Even with filtered events, this is still necessary to prevent acting on events raised by scripts!
    -- (This fixes a bug with Rocket-Silo-Construction where the silo would be removed after the first
    -- stage was finished.)
    -- Event filtering works correctly since Factorio 0.18.27. We still keep this in for now so the same
    -- code can be used with Factorio 0.17.79 as long as it is the stable version!
    if (not entity.valid) or entity.name ~= "car-key" then
        GCKI.dprint("Leaving event script on_built_entity early (nothing to do for entity \"" ..
                        entity.name .. "\")")
        return
    end
    ------------------------------------------------------------------------------------

    -- In editor or god mode, there is no character, so the game would crash if we continued.
    if not player.character then
        entity.destroy()
        player.cursor_stack.set_stack{name = "car-key", count = 1}
        GCKI.output(player, {"GCKI-messages.editor_or_god_mode"})
       return
    end

    --~ local owner = nil                                   -- Player who owns a vehicle
    --~ local locker = nil                                  -- Player who locked a vehicle
    --~ local owned_by = owned_by_player(player)        -- Vehicle owned by a player
    --~ local occupants = nil                               -- Players who are driver/passenger in a vehicle

    -- We want to wait some time between actions, but we may need to reset vehicle color to the
    -- color of the owner (or the locker if there is no owner)!
    --~ for owner, vehicle in pairs(global.vehicles) do
        --~ -- Check whether vehicle is valid -- prevents crashes if a vehicle was from an deactived/removed mod
        --~ if vehicle.valid then
            --~ vehicle.color = game.players[owner].color
        --~ else
            --~ GCKI.dprint("Vehicle doesn't exist -- perhaps it's from an inactive mod?")
        --~ end
    --~ end
    for v, vehicle in pairs(global.GCKI_vehicles) do
        color_vehicle(vehicle.entity)
    end

    if global.ticks_since_action < 60 then
        entity.destroy()
        player.cursor_stack.set_stack{name = "car-key", count = 1}
        ------------------------------------------------------------------------------------
        -- NEW: If we just add a car-key to player.cursor_stack, we will end up with one
        --      additional key, so we also remove one key from the player's inventory.
        if player.get_main_inventory().get_item_count("car-key") > 0 then
            player.remove_item({name = "car-key", count = 1})
        end
        ------------------------------------------------------------------------------------
        GCKI.dprint("Leaving event script on_built_entity for player " ..
                --~ game.players[event.player_index].name .. " early (global.ticks_since_action:\t" ..
                game.players[index].name .. " early (global.ticks_since_action:\t" ..
                global.ticks_since_action .. " < 60: " .. tostring(global.ticks_since_action < 60) .. ")")
        return
    end

    global.ticks_since_action = 0

    local owner = nil                                   -- Player who owns a vehicle
    local locker = nil                                  -- Player who locked a vehicle
    local owned_by = owned_by_player(player)        -- Vehicle owned by a player
    local occupants = nil                               -- Players who are driver/passenger in a vehicle

    if not player.cursor_stack.valid_for_read then
        player.cursor_stack.set_stack{name = "car-key", count = 1}
        ------------------------------------------------------------------------------------
        -- NEW: If we just add a car-key to player.cursor_stack, we will end up with one
        --      additional key, so we also remove one key from the player's inventory.
        if player.get_main_inventory().get_item_count("car-key") > 0 then
            player.remove_item({name = "car-key", count = 1})
        end
        ------------------------------------------------------------------------------------
    elseif player.cursor_stack.name == "car-key" then
        player.cursor_stack.count = player.cursor_stack.count + 1
        ------------------------------------------------------------------------------------
        -- NEW: Removed item spilling! This shouldn't be needed in Factorio 0.17 as the
        --      inventory slot is now reserved if the player has moved a complete stack
        --      from the inventory to the cursor.
    --~ elseif player.insert({name = "car-key", count = 1}) == 0 then
        --~ player.surface.spill_item_stack(player.position,{name = "car-key", count = 1})
        ------------------------------------------------------------------------------------
    end

    -- Get the location the key was used and find any vehicles in range
    local vehicle = nil
    local vehicles = entity.surface.find_entities_filtered{area = area_create(entity.position,1), type = "car"}

    -- Loop over all the objects we found and find the first vehicle
    for i, _ in pairs(vehicles) do
        local find_vehicle = vehicles[i]
        if find_vehicle.type == "car" then
            -- Dont save turrets (for compatibility with the "Vehicle turrets" mod)
            local wordIsThere = string.find(tostring(find_vehicle.name), "turret")
            if not wordIsThere then
                vehicle = find_vehicle
                break
            end
        end
    end

    -- Found vehicle
    if vehicle then
        owner = is_owned_vehicle(vehicle)
        occupants = GCKI.get_occupants(vehicle)

        -- Player tried to take our vehicle
        if owner and not (player == owner) then
            -- Message to player: "This [vehicle] is owned by [owner]."
            GCKI.output(player, {"GCKI-messages.owner", GCKI.loc_name(vehicle), owner.name})
            -- Message to owner: "[Player] has tried to take your [vehicle]!"
            GCKI.output(owner, {"GCKI-messages.thief", player.name, GCKI.loc_name(vehicle)})
        end
        -- Vehicle is not owned by anybody else
        if not owner or (owner == player) then
            -- You own this vehicle, lock or unlock the doors
            if player == owner then
                locker = is_locked_vehicle(vehicle)
GCKI.dprint("Player is owner. Vehicle is locked: ", tostring(locker))
GCKI.show(vehicle.active, "v.active")
GCKI.show(vehicle.minable, "v.minable")
GCKI.show(vehicle.operable, "v.operable")
                lock_or_unlock(vehicle, player)
GCKI.dprint("Returned from lock_or_unlock")
GCKI.show(vehicle.active, "v.active")
GCKI.show(vehicle.minable, "v.minable")
GCKI.show(vehicle.operable, "v.operable")
                local locked = {"GCKI-states.locked"}
                -- Vehicle has been unlocked
                if locker then
                    locked = {"GCKI-states.unlocked"}
                    remove_vehicle_locker(vehicle, player)
                -- Vehicle has been locked. Eject all occupants!
                else
                    add_vehicle_locker(vehicle, player)
                    --~ ------------------------------------------------------------------------------------
                    --~ --        COMPATIBILITY WITH MODS THAT PLACE DUMMY CHARACTERS IN VEHICLES         --
                    --~ -- Remove dummy characters inserted by other mods so they don't spam the surface. --
                    --~ ------------------------------------------------------------------------------------
                    --~ local dummy = is_dummy_inside(vehicle)
                    --~ if dummy then
                        --~ for role, character in pairs(dummy) do
                            --~ GCKI.dprint({"", "Removing dummy ", role, " ", tostring(character and
                                --~ character.name), " from ", GCKI.loc_name(vehicle), " (on_built_entity)!"})

                            --~ character.destroy()
                            --~ occupants[role] = nil
                            --~ dummy[role] = nil
                        --~ end
                    --~ end
                    ------------------------------------------------------------------------------------
                    --               FIX FOR AAI PROGRAMMABLE VEHICLES AND SIMILAR MODS               --
                    ------------------------------------------------------------------------------------
                    remove_dummy(vehicle)
                    ------------------------------------------------------------------------------------
                    vehicle.set_passenger(nil)
                    vehicle.set_driver(nil)
                    -- Inform occupants that vehicle has been locked
                    if occupants then
                        for _, occupant in pairs(occupants) do
                            -- Player will always be informed
                            if player == occupant then
                                GCKI.output(occupant, {"GCKI-messages.lock_vehicle_eject_owner",
                                                            GCKI.loc_name(vehicle)})
                            --~ -- Other occupants be informed unless they are a dummy passenger inserted
                            --~ -- by the Autodrive mod
                            --~ elseif dummy ~= occupant then
                            -- Other occupants will be informed unless they are a dummy passenger
                            -- inserted by another mod
                            else
                                GCKI.output(occupant, {"GCKI-messages.lock_vehicle_eject_other",
                                                            player.name, GCKI.loc_name(vehicle)})
                            end
                        end
                    end
                end
                -- Color the vehicle with owner's color
                color_vehicle(vehicle)

                --~ GCKI.output(player, {"GCKI-messages.lock", GCKI.loc_name(global.vehicles[index]), locked})
                GCKI.output(player, {"GCKI-messages.lock", GCKI.loc_name(vehicle), locked})
            -- Vehicle has no owner, but may be locked because the last owner claimed another
            else
                -- Only unlocked vehicles can be claimed
                GCKI.dprint("Check if " .. vehicle.name .. " is locked:")
                locker = is_locked_vehicle(vehicle)

                if not locker or (player == locker) then
                    if vehicle.type == "car" then
                        -- Dont save turrets
                        -- (I guess that is for compatibility with the "Vehicle turrets" mod)
                        local wordIsThere = string.find(tostring(vehicle.name), "turret")
                        if not wordIsThere then
                            ------------------------------------------------------------------------------------
                            -- NEW: Vehicle has no owner -- but we can't claim it if somebody else is inside!
                            local occupied = false
                            -- Vehicle must have no passenger, and only the driver may claim it !
                            if occupants and (
                                (occupants["passenger"] and occupants["passenger"].is_player()) or
                                (occupants["driver"].is_player() and occupants["driver"] ~= player)
                            ) then
                                occupied = true
                            end
                            -- Vehicle has passenger or player is not the driver
                            if occupied then
                                GCKI.output(player, {"", {"GCKI-messages.vehicle_occupied",
                                                GCKI.loc_name(vehicle), get_occupant_names(vehicle)}, " ",
                                                {"GCKI-messages.vehicle_find_other"}})
                            -- This vehicle belongs to us now!
                            else
                                -- Player owned and locked this vehicle before abandoning it. Remind him to unlock it.
                                if player == locker then
                                    GCKI.output(player,  {"", {"GCKI-messages.new_vehicle", GCKI.loc_name(vehicle)}, " ",
                                            {"GCKI-messages.lock_state", GCKI.loc_name(vehicle), {"GCKI-states.locked"}}})
                                -- Inform player that this vehicle belongs to him now.
                                else
                                    GCKI.output(player, {"GCKI-messages.new_vehicle", GCKI.loc_name(vehicle)})
                                end
                                -- Did the player own another vehicle before claiming this one?
                                --~ local old_vehicle_id = owned_by_player(player)
                                local old_vehicle_id = owned_by
                                if old_vehicle_id then
--~ GCKI.dprint("Player " .. serpent.line(player and player.name) .. " has claimed a new vehicle. Removing " .. serpent.line(player and player.name) .. " as owner of " .. old_vehicle_id)
                                    local old_vehicle = global.GCKI_vehicles[old_vehicle_id].entity
                                    -- Remove owner from vehicle
                                    remove_vehicle_owner(old_vehicle_id, player)
                                    color_vehicle(old_vehicle)
                                end
                                -- Mark vehicle as owned by player
                                --~ global.GCKI_vehicles[vehicle.unit_number].owner = index
                                add_vehicle_owner(vehicle, player)
                                -- Color the (new) vehicle with player color
                                color_vehicle(vehicle)
                            end
                        end
                    end
                else
                GCKI.output(player, {"GCKI-messages.vehicle_locked", GCKI.loc_name(vehicle), locker.name})
                end
            end
        end
    -- Key was used, but not on a vehicle -- player tried to summon his vehicle
    --~ elseif global.vehicles[index].valid then
    elseif global.GCKI_vehicles[owned_by] and
           global.GCKI_vehicles[owned_by].entity and
           global.GCKI_vehicles[owned_by].entity.valid then
                --Get current vehicle
        --~ local vehicle = global.vehicles[index]
        local vehicle = global.GCKI_vehicles[owned_by].entity
        -- Only summon vehicle if it really exists
        if vehicle.type == "car" then
            -- Check if vehicle is locked
            local locker = is_locked_vehicle(vehicle)
            ------------------------------------------------------------------------------------
            -- NEW: Replaced original code -- vehicle is teleported now instead of copied/replaced
            -- Check whether selected position is in player's reach
            local max_distance = 0
            if GCKI.long_reach then
                max_distance = (player.reach_distance)
            else
                max_distance = (player.reach_distance - player.character_reach_distance_bonus)
            end
            --~ GCKI.dprint("Player reach distance: " .. tostring(max_distance))
            local distance = get_distance(player.position, entity.position)

            -- Check whether chunk where vehicle should be placed has been generated yet
            local is_valid_chunk = entity.surface.is_chunk_generated({entity.position.x/32, entity.position.y/32})

            -- Check whether vehicle can be placed at the selected position
            local is_placeable = entity.surface.can_place_entity{
                name=vehicle.name,
                position=entity.position,
                direction=vehicle.direction,
                force=player.force,
                build_check_type=defines.build_check_type.manual,
                forced=false
            }

            -- Teleport vehicle to new position
            if is_valid_chunk and is_placeable and not (distance > max_distance) then
                local occupants = GCKI.get_occupants(vehicle)
                -- Teleport on same surface
                if entity.surface == vehicle.surface then
                    vehicle.teleport(entity.position, entity.surface)
                    -- Inform owner that his vehicle has arrived
                    GCKI.output(player, {"GCKI-messages.arrived", GCKI.loc_name(vehicle)})
                    -- Inform driver/passenger that they have been summoned with the vehicle to its owner
                    if occupants then
                        for _, test in pairs(occupants) do
                            -- Only inform other occupants that the vehicle has been summoned;
                            -- if the owner/summoner is inside, he already knows he summoned his vehicle!
                            if test and not (player == test) then
                                GCKI.output(test, {"GCKI-messages.summoned", GCKI.loc_name(vehicle), player.name})
                            end
                        end
                    end
                -- Teleport to another surface
                else
                    --Does the vehicle have driver or passenger?
                    if occupants then
                        -- Do the occupants want to be moved to other surfaces?
                        for role, occupant in pairs(occupants) do
                            -- Owner can't summon himself to another surface, so we don't need to
                            -- check for "player == occupant" before informing occupants that they've
                            -- been summoned to another surface.
                            --~ if global.player_settings[occupant.index].teleport_to_other_surface then
                            if not occupant.is_player() then
GCKI.dprint("Teleporting dummy " .. role .. occupant.name " to " .. entity.surface.name ..".")
                            elseif global.player_settings[occupant.index].teleport_to_other_surface then
                                -- Inform driver/passenger that they've been moved to another surface
                                GCKI.output(occupant, {"",
                                        {"GCKI-messages.summoned_other_surface", player.name,
                                                GCKI.loc_name(vehicle), entity.surface.name}, " ",
                                        {"GCKI-messages.summoned_other_surface_moved", GCKI.loc_name(vehicle)}, " ",
                                        {"GCKI-messages.summoned_check_surface_setting",
                                                {"mod-setting-name.GCKI-teleport_to_other_surface"}}
                                    })
                            -- Inform occupant he's about to be ejected
                            else
                                GCKI.output(occupant,{"",
                                    {"GCKI-messages.summoned_other_surface", player.name, GCKI.loc_name(vehicle),
                                        entity.surface.name}, " ",
                                    {"GCKI-messages.summoned_other_surface_eject", vehicle.surface.name}, " ",
                                    {"GCKI-messages.summoned_check_surface_setting",
                                            {"mod-setting-name.GCKI-teleport_to_other_surface"}}
                                    })
                                -- Eject driver
                                if role == "driver" then
                                    vehicle.set_driver(nil)
                                -- Eject passenger
                                else
                                    vehicle.set_passenger(nil)
                                end
                            end
                        end
                    end
                    -- Get unit_number of vehicle so we can find it again after it was teleported
                    local vehicle_id = vehicle.unit_number
                    -- Successfully teleported vehicle to other surface
                    if vehicle.teleport(entity.position, entity.surface) then
                        -- Vehicle isn't valid after teleporting, so we need to find it again on the new
                        -- surface. We can't check for entity name because mods may have changed it, so we
                        -- filter by entity type.
                        vehicles = entity.surface.find_entities_filtered{type="car", position=entity.position}
                        -- Return value is an array, so identify the teleported vehicle. (If other mods remove the collision of vehicles, several may be stacked in the same position.)
                        for _,v in pairs(vehicles) do
                            if vehicle_id == v.unit_number then
                                -- Set vehicle to found vehicle
                                vehicle = v
                                -- If vehicle is locked, update entry in list of locked vehicles
                                if locker then
                                    -- Remove old vehicle (now unvalid) from list of locked vehicles
                                    remove_invalid_locked_vehicles(player)
                                    -- Add new vehicle to list of locked vehicles
                                    add_vehicle_locker(vehicle, player)
                                end
                                -- Update entry in list of owned vehicles
                                --~ global.vehicles[index] = vehicle
                                global.GCKI_vehicles[vehicle.unit_number].locker = index
                                break
                            end
                        end
                        -- Inform owner that his vehicle has arrived
                        GCKI.output(player, {"GCKI-messages.arrived", GCKI.loc_name(vehicle)})
                    -- This should never be triggered!
                    else
                        GCKI.dprint("Couldn't teleport " .. vehicle.name ..  " to surface " .. entity.surface.name .. "!")
                    end
                end
            -- New position is within player's reach, but vehicle can't be placed there
            elseif (not is_valid_chunk) or (not is_placeable) then
                GCKI.output(player, {"GCKI-messages.wrong_place", GCKI.loc_name(vehicle)})
            -- Player can't reach the new position
            else
                GCKI.output(player, {"GCKI-messages.far_place", GCKI.loc_name(vehicle)})
            end
        ------------------------------------------------------------------------------------
        -- Selected vehicle is not of type "car"
        else
            GCKI.output(player, {"GCKI-messages.invalid"})
        end
    -- Player doesn't own a vehicle yet
    else
        GCKI.output(player, {"GCKI-messages.vehicle_not_yet"})
    end
    entity.destroy()

    GCKI.dprint("End of event script on_built_entity for player " .. game.players[event.player_index].name)
-- Event filtering doesn't work for events raised by scripts, so we need to include filtering code
-- anyway. Let's just remove event filtering for now -- then the mod can also be used with the current stable
-- version of Factorio (0.17.74 as of 2019-11-19).
end, { {filter = "name", name = "car-key"} })
--~ end)

------------------------------------------------------------------------------------
--                         DRIVING STATE OF A PLAYER CHANGED                      --
------------------------------------------------------------------------------------
script.on_event(defines.events.on_player_driving_changed_state, function(event)
    GCKI.dprint("Entered event script on_player_driving_changed_state for " ..
                                    tostring(game.players[event.player_index].name))
    local player_index = event.player_index
    local player = game.players[player_index]
GCKI.dprint("Player: " .. player.name or "no name")
    -- In editor or god mode, there is no character, so the game would crash if we continued.
    if not player.character then
        GCKI.dprint("Leaving event script on_player_driving_changed_state early" ..
                        "(nothing to do in editor mode or god mode).")
        return
    end


    local vehicle = event.entity

    -- Check whether the vehicle still exists before trying to act on it!
    if not (vehicle and vehicle.valid) then
        GCKI.dprint("Vehicle has been destroyed while the player was in it!")
    -- Only act if an entity of prototype "car" been entered/left
    elseif vehicle.type == "car" then
        -- We want to wait some time between actions, but we may need to reset vehicle color to owner color!
        color_vehicle(vehicle)

        -- Set the gunner of the vehicle
        set_shooter(vehicle)

        -- Add delay between actions
        if global.ticks_since_action < 60 then
            GCKI.dprint("Left on_player_driving_changed_state early (global.ticks_since_action:\t" ..
                            global.ticks_since_action .. " < 60: " .. tostring(global.ticks_since_action < 60) .. ")")
            return
        end

        global.ticks_since_action = 0

        --~ local occupants = GCKI.get_occupants(vehicle)        -- Players who drive in a vehicle
        --~ local owner = is_owned_vehicle(vehicle)
        --~ local locker = is_locked_vehicle(vehicle)            -- Player who locked a vehicle
--~ GCKI.dprint("Locker: " .. tostring(locker and locker.name or "no locker"))
--~ GCKI.dprint("owner: " .. tostring(is_owned_vehicle(vehicle) and is_owned_vehicle(vehicle).name or "no  owner"))

        --~ ------------------------------------------------------------------------------------
        --~ --        COMPATIBILITY WITH MODS THAT PLACE DUMMY CHARACTERS IN VEHICLES         --
        --~ -- Remove dummy characters inserted by other mods so they don't spam the surface  --
        --~ -- if they are ejected, and so the player can get in the driver's seat.           --
        --~ ------------------------------------------------------------------------------------
--~ GCKI.dprint("Driving state changed: check for dummies!")
        --~ local dummy = is_dummy_inside(vehicle)

--~ GCKI.dprint("Occupants: " .. serpent.block(GCKI.get_occupants(vehicle)))
        --~ if dummy then
--~ GCKI.dprint("Found dummy")
            --~ for role, character in pairs(dummy) do
                --~ GCKI.dprint({"", "Removing dummy ", role, " ", tostring(character and
                    --~ character.name), " from ", GCKI.loc_name(vehicle), " (on_player_driving_changed_state)!"})

                --~ character.destroy()
                --~ occupants[role] = nil
                --~ dummy[role] = nil
            --~ end
        --~ end
--~ GCKI.dprint("Occupants after removal of dummies: " .. serpent.block(GCKI.get_occupants(vehicle)))
        ------------------------------------------------------------------------------------
        --               FIX FOR AAI PROGRAMMABLE VEHICLES AND SIMILAR MODS               --
        ------------------------------------------------------------------------------------
        remove_dummy(vehicle)

        local occupants = GCKI.get_occupants(vehicle)        -- Players who drive in a vehicle
        local owner = is_owned_vehicle(vehicle)              -- Player who owns a vehicle
        local locker = is_locked_vehicle(vehicle)            -- Player who locked a vehicle
GCKI.dprint("Locker: " .. tostring(locker and locker.name or "no locker"))
GCKI.dprint("owner: " .. tostring(is_owned_vehicle(vehicle) and is_owned_vehicle(vehicle).name or "no  owner"))

        -- Dummy was driver and player is passenger? Make player driver!
GCKI.dprint("not driver: " .. tostring(occupants and not occupants["driver"]))
GCKI.dprint("player == passenger: " .. tostring(player == (occupants and occupants["passenger"])))
        if occupants and player == occupants["passenger"] and not occupants["driver"] then
            vehicle.set_driver(player)
        end
local test = GCKI.get_occupants(vehicle)
GCKI.dprint("Occupants after changing driver: " .. serpent.block(test))
--~ if test then
--~ for k,v in pairs(test) do GCKI.dprint(k .. ": " .. v.name) end
--~ end
        ------------------------------------------------------------------------------------

        -- Vehicle has an owner or is locked
        if owner or locker then
            -- Vehicle is owned by player
            if owner == player then
                GCKI.dprint({"", "This is your own ", GCKI.loc_name(vehicle), "!"})

                -- We can't enter a locked vehicle!
                if locker then
                    vehicle.set_passenger(nil)
                    vehicle.set_driver(nil)
                    GCKI.output(player, {"GCKI-messages.lock_vehicle_enter_owner", GCKI.loc_name(vehicle)})
                    -- Color the vehicle with owner's color
                    color_vehicle(vehicle)
                -- This is our vehicle -- we want to be the driver!
                else
                    if occupants then
--~ GCKI.dprint("We want to be driver! Occupants: ")
--~ GCKI.dprint("Driver: " .. tostring(occupants["driver"] and occupants["driver"].name))
--~ GCKI.dprint("Passenger: " .. tostring(occupants["passenger"] and occupants["passenger"].name))
                        -- We couldn't have entered a vehicle that already had driver and passenger, so if
                        -- the vehicle already has a driver, we can safely replace him and make him passenger.
                        if (player == occupants["passenger"]) then
                            -- Make owner driver
                            vehicle.set_driver(player)
--~ ------------------------------------------------------------------------------------
--~ -- Inform the player that he entered his own vehicle
--~ -- This is only needed until Factorio 0.17.70, when vehicles will always have the
--~ -- color of the owner!
--~ GCKI.output(owner, {"GCKI-messages.unlock_vehicle_enter_owner", GCKI.loc_name(vehicle)})
--~ ------------------------------------------------------------------------------------

                            -- Only act for real players, ignore dummy characters!
                            if occupants["driver"] and
                               occupants["driver"].valid and
                               occupants["driver"].is_player() then

                                GCKI.output(player, {"", {"GCKI-messages.new_occupant_role",
                                                         {"GCKI-words.driver"},
                                                         GCKI.loc_name(vehicle)}, " ",
                                                         {"GCKI-messages.new_occupant_role_other",
                                                            occupants["driver"].name, {"GCKI-words.passenger"}}
                                                    })
                                -- Make previous driver passenger
                                vehicle.set_passenger(occupants["driver"])
                                GCKI.output(occupants["driver"], {"",
                                                {"GCKI-messages.vehicle_owner_is_driver",
                                                    GCKI.loc_name(vehicle), owner.name}, " ",
                                                {"GCKI-messages.new_occupant_role", {"GCKI-words.passenger"},
                                                    GCKI.loc_name(vehicle)}
                                })
                            end
                        -- Player is owner and driver
                        elseif player == occupants["driver"] then
--~ ------------------------------------------------------------------------------------
--~ -- Inform the player that he entered his own vehicle
--~ -- This is only needed until Factorio 0.17.70, when vehicles will always have the
--~ -- color of the owner!
--~ GCKI.output(owner, {"GCKI-messages.unlock_vehicle_enter_owner", GCKI.loc_name(vehicle)})
--~ ------------------------------------------------------------------------------------
                            if occupants["passenger"] then
                                GCKI.output(occupants["passenger"], {"GCKI-messages.vehicle_owner_is_driver",
                                            GCKI.loc_name(vehicle), owner.name})
                            end
                        -- Player is owner and has left
                        else
                            -- Inform owner
                            local occupants = GCKI.get_occupants(vehicle)
                            local msg = {"GCKI-messages.owner_left_own", GCKI.loc_name(vehicle)}
                            if occupants then
                                msg = {"", {"GCKI-messages.owner_left_own", GCKI.loc_name(vehicle)}, " ",
                                           {"GCKI-messages.vehicle_occupied", GCKI.loc_name(vehicle),
                                                get_occupant_names(vehicle)}}
                            end
                            GCKI.output(owner, msg)

                            -- Inform remaining occupants
                            for _, player in pairs(occupants) do
                                GCKI.output(player, {"GCKI-messages.owner_left", owner.name, GCKI.loc_name(vehicle)})
                            end
                        end
                    end
                end
            -- Vehicle is owned by another player or has no owner
            else
                GCKI.dprint({"", "This ", GCKI.loc_name(vehicle), " belongs to ", tostring(owner and owner.name or "nobody"), "!"})
                -- If vehicle is locked, nobody may enter it!
                if locker then
                    vehicle.set_passenger(nil)
                    vehicle.set_driver(nil)
                    -- Color the vehicle with locker's color
                    color_vehicle(vehicle)
                    -- If vehicle has driver/passenger, inform them that the vehicle has been locked.
                    if occupants then
                        for _, occupant in pairs(occupants) do
                            -- Only output message to real players
                            if occupant.is_player() then
                                if player == locker then
                                GCKI.output(occupant, {"GCKI-messages.lock_vehicle_enter_locker",
                                                        GCKI.loc_name(vehicle)})
                                else
                                GCKI.output(occupant, {"GCKI-messages.lock_vehicle_enter_other",
                                                        GCKI.loc_name(vehicle), locker.name})
                                end
                            -- Remove dummy characters
                            else
                                occupant.destroy()
                            end
                        end
                    end
                -- Vehicle is not locked and does not belong to player.
                -- If the vehicle is owned by somebody else, the player may drive in it, but should be aware
                -- that the owner can summon the vehicle anytime and anywhere he likes. Also, the owner
                -- should be told that somebody has stolen his vehicle.
                else
                    -- Vehicle has driver and/or passenger and is unlocked
                    if occupants then
                    -- Player has entered the vehicle and is driver
                        if player == occupants["driver"] then
                            -- Inform owner that somebody entered the car
                            if owner then
                                local msg = {"GCKI-messages.vehicle_unlock_warn_owner", player.name,
                                                    {"GCKI-words.driver"}, GCKI.loc_name(vehicle)}
                                -- Inform owner who else is in the vehicle
                                if occupants["passenger"] then
                                    msg = {"", {"GCKI-messages.vehicle_unlock_warn_owner", player.name,
                                                    {"GCKI-words.driver"}, GCKI.loc_name(vehicle)}, " ",
                                               {"GCKI-messages.vehicle_occupied",
                                                    GCKI.loc_name(vehicle), get_occupant_names(vehicle)}
                                    }
                                end
                                -- Don't print message if owner is not at least 1 chunk away from the vehicle
                                if get_distance(owner.position, vehicle.position) > message_owner_distance then
                                    GCKI.output(owner, msg)
                                end
                            end
                            -- Inform occupants
                            for role, player in pairs(occupants) do
                                -- Tell the driver whether he has a passenger
                                if occupants["passenger"] then
                                    GCKI.output(player, {"GCKI-messages.vehicle_occupied",
                                                        GCKI.loc_name(vehicle), get_occupant_names(vehicle)})
                                end
                                -- Tell players their new role
                                local msg = ""
                                if role == "driver" then
                                    msg = {"GCKI-messages.new_occupant_role", {"GCKI-words.driver"},
                                                GCKI.loc_name(vehicle)}
                                else
                                    msg = {"GCKI-messages.new_occupant_role", {"GCKI-words.passenger"},
                                                GCKI.loc_name(vehicle)}
                                end
                                GCKI.output(player, msg)
                                -- Warn players that the vehicle may be summoned by its owner
                                GCKI.output(player, {"GCKI-messages.vehicle_unlock_warn_guest",
                                                GCKI.loc_name(vehicle), owner.name})
                            end
                        -- Player has entered the vehicle and is passenger
                        elseif player == occupants["passenger"] then
                            -- Inform vehicle owner that his vehicle has a new passenger
                            if owner and (get_distance(owner.position, vehicle.position) > message_owner_distance) then
                                GCKI.output(owner, {"GCKI-messages.unlock_vehicle_enter_other", player.name,
                                                    GCKI.loc_name(vehicle), {"GCKI-words.passenger"}})
                            end
                            -- Inform driver that he has a passenger (unless driver is the owner of this vehicle)
                            if occupants["driver"] and not (occupants["driver"] == owner) then
                                GCKI.output(occupants["driver"], {"GCKI-messages.new_occupant_role_other",
                                                    player.name, {"GCKI-words.passenger"}})
                            end
                            -- Tell player who is the driver
                            -- Owner is driver
                            if occupants["driver"] == owner then
                                GCKI.output(player, {"GCKI-messages.vehicle_owner_is_driver",
                                                    GCKI.loc_name(vehicle), owner.name})
                            -- Somebody else is driver
                            elseif occupants["driver"] then
                                GCKI.output(player, {"GCKI-messages.new_occupant_role_other",
                                                    occupants["driver"].name, {"GCKI-words.driver"}})
                            -- Vehicle has no driver
                            -- (Note to self: Is that even possible?!)
                            else
                                GCKI.output(player, {"GCKI-messages.vehicle_no_driver", GCKI.loc_name(vehicle)})
                            end
                            -- Tell player that he is the passenger
                            GCKI.output(player, {"GCKI-messages.new_occupant_role",
                                                {"GCKI-words.passenger"}, GCKI.loc_name(vehicle)})
                            -- Warn player that vehicle may be summoned by owner (unless owner is the driver)
                            if not occupants["driver"] == owner then
                                GCKI.output(player, {"GCKI-messages.vehicle_unlock_warn_guest",
                                                    GCKI.loc_name(vehicle), owner.name})
                            end

                        -- Player has left the vehicle
                        else
                            -- Color the vehicle with owner's color
                            color_vehicle(vehicle)
                            -- Inform vehicle owner if he's more than a chunk away from the vehicle
                            if get_distance(owner.position, vehicle.position) > message_owner_distance then
                                GCKI.output(owner, {"GCKI-messages.passenger_left",
                                                    player.name, GCKI.loc_name(vehicle)})
                            end
                            -- If vehicle is empty, tell player he can steal it!
                            if not occupants then
                                GCKI.output(player, {"GCKI-messages.vehicle_steal",
                                                GCKI.loc_name(vehicle), owner.name})
                            -- If player was driver, inform passenger
                            elseif occupants["passenger"] and not (player == occupants["passenger"]) then
                                GCKI.output(occupants["passenger"],
                                            {"GCKI-messages.occupant_left", player.name,
                                             GCKI.loc_name(vehicle), {"GCKI-words.driver"}})
                            -- If player was passenger, inform driver
                            elseif occupants["driver"] and not (player == occupants["driver"]) and
                                    not (owner == occupants["driver"]) then
                                GCKI.output(occupants["driver"],
                                            {"GCKI-messages.occupant_left", player.name,
                                             GCKI.loc_name(vehicle),
                                             {"GCKI-words.passenger"}})
                            end
                        end

                    -- Vehicle has an owner and is unlocked; player has left the vehicle
                    else
                        -- Color the vehicle with owner color
                        color_vehicle(vehicle)
                        -- Inform vehicle owner if he's more than a chunk away from the vehicle
                            if get_distance(owner.position, vehicle.position) > message_owner_distance then
                                GCKI.output(owner, {"GCKI-messages.passenger_left", player.name,
                                                        GCKI.loc_name(vehicle)})
                            end
                        -- If vehicle is empty, tell player he can steal it!
                        if not occupants then
                            GCKI.output(player, {"GCKI-messages.vehicle_steal",
                                               GCKI.loc_name(vehicle), owner.name})
                        -- Inform the remaining occupant of the vehicle if driver or passenger has left
                        else
                            -- If player was driver, inform passenger
                            if  occupants["passenger"] and not (player == occupants["passenger"]) then
                                GCKI.output(occupants["passenger"],
                                            {"GCKI-messages.occupant_left", player.name,
                                                GCKI.loc_name(vehicle), {"GCKI-words.driver"}})
                            -- If player was passenger, inform driver
                            elseif occupants["driver"] and not (player == occupants["driver"]) and
                                    not (owner == occupants["driver"]) then
                                GCKI.output(occupants["driver"], {"GCKI-messages.occupant_left", player.name,
                                                GCKI.loc_name(vehicle), {"GCKI-words.passenger"}})
                            end
                        end
                    end
                end
            end
        -- Vehicle has no owner and is unlocked
        else
            -- Player has entered the vehicle, both driver and passenger are present
            if occupants and occupants["driver"] and occupants["passenger"] then
                -- Inform occupants
                for role, player in pairs(occupants) do
--~ GCKI.dprint("Driver: " .. player.name .. " (" .. tostring(type(player)) .. ")")
--~ GCKI.dprint("Passenger: " .. player.name .. " (" .. tostring(type(player)) .. ")")

                    -- Tell player who is in the vehicle
                    GCKI.output(player, {"GCKI-messages.vehicle_occupied",
                                            GCKI.loc_name(vehicle), get_occupant_names(vehicle)})
                    -- Tell player his new role
                    if role == "driver" then
                        GCKI.output(player, {"GCKI-messages.new_occupant_role",
                                                {"GCKI-words.driver"}, GCKI.loc_name(vehicle)})
                    else
                        GCKI.output(player, {"GCKI-messages.new_occupant_role", {"GCKI-words.passenger"},
                                    GCKI.loc_name(vehicle)})
                    end
                end
            -- Player has left vehicle, either driver or passenger are present
            elseif occupants then
                -- Driver has left
                if occupants["passenger"] and not (player == occupants["passenger"]) then
                    GCKI.output(occupants["passenger"], {"GCKI-messages.occupant_left",
                        player.name, GCKI.loc_name(vehicle), {"GCKI-words.driver"}})
                -- Passenger has left
                elseif occupants["driver"] and not (player == occupants["driver"]) then
                    GCKI.output(occupants["driver"], {"GCKI-messages.occupant_left",
                        player.name, GCKI.loc_name(vehicle), {"GCKI-words.passenger"}})
                end
            -- Nobody is in the vehicle
            else
                -- Reset vehicle color to show it's unowned and without occupant
                color_vehicle(vehicle)
                -- Inform player that the vehicle can be claimed
                GCKI.output(player, {"GCKI-messages.vehicle_available", GCKI.loc_name(vehicle)})
            end
        end
    set_shooter(vehicle)
    end
    GCKI.dprint("End of event script on_player_driving_changed_state for " ..
                                    tostring(game.players[event.player_index].name))
end)

------------------------------------------------------------------------------------
--                             ADD REMOTE INTERFACES                              --
------------------------------------------------------------------------------------
remote.add_interface('GCKI', {
    -- Returns player who locked the vehicle or nil if it's unlocked
    vehicle_locked_by = function(vehicle)
        return is_locked_vehicle(vehicle)
    end,

    -- Returns player who owns the vehicle or nil if it has no owner
    vehicle_owned_by = function(vehicle)
        return is_owned_vehicle(vehicle)
    end,

    -- Returns unit_number of vehicle owned by the player, or nil
    owned_by_player = function(player)
        return owned_by_player(player)
    end,

    -- Returns dummy characters in a vehicle or nil
    dummies_in_vehicle = function(vehicle)
        return is_dummy_inside(vehicle)
    end,

    -- For mods like AAI that replace vehicles
    -- (add new vehicle to list immediately after old one has been removed)
    on_entity_replaced = function(event)
        return replace_vehicle(event)
    end,

    -- Make data GCKI stores with a vehicle available to other mods. Make that
    -- general, so other mods can use this even if we add some new data in the future!
    get_vehicle_data = function(vehicle)
        local id = (type(vehicle) == "number" and vehicle) or
                   (vehicle and vehicle.valid and vehicle.type == "car" and vehicle.unit_number) or
                   GCKI.error("Invalid value for vehicle: " .. GCKI.argprint(vehicle))

        local v = global.GCKI_vehicles[id]
        local data = {}
        if v then
            data.owner = v.owner
            data.locker = v.locker
        end

        return (table_size(data) > 0) and data or nil
    end,

    -- For mods like Vehicle Wagons that temporarily remove a vehicle and
    -- restore it after an indeterminate time)
    vehicle_removed = function(vehicle, modname)
GCKI.dprint("GCKI: Entered interface function vehicle_removed")
        local id = (type(vehicle) == "number" and vehicle) or
                   (vehicle and vehicle.valid and vehicle.type == "car" and vehicle.unit_number) or
                   GCKI.error("Invalid value for vehicle: " .. GCKI.argprint(vehicle))
        local v = global.GCKI_vehicles[id]
GCKI.dprint("vehicle: " .. serpent.block(vehicle.name))
GCKI.dprint("modname: " .. serpent.block(modname))
        if v and v.entity.valid and v.owner then
            GCKI.output(game.players[v.owner], {"GCKI-messages.vehicle_unavailable", GCKI.loc_name(v.entity)})
        end
        remove_vehicle(id)
GCKI.dprint("GCKI: End of interface function vehicle_removed")
    end,
    --~ vehicle_restored = function(new_vehicle, data, modname)
--~ GCKI.dprint("GCKI: Entered interface function vehicle_restored")
--~ GCKI.dprint("new_vehicle: " .. serpent.block(new_vehicle.name))
--~ GCKI.dprint("data: " .. serpent.block(data))
--~ GCKI.dprint("modname: " .. serpent.block(modname))

        --~ restore_vehicle(new_vehicle, data)
--~ GCKI.dprint("GCKI: End of interface function vehicle_restored")
   --~ end,

    -- Reset vehicle color
    reset_vehicle_color = function(vehicle, autodrive_owner)
        color_vehicle(vehicle, autodrive_owner)
    end,
})

------------------------------------------------------------------------------------
--                    FIND LOCAL VARIABLES THAT ARE USED GLOBALLY                 --
--                              (Thanks to eradicator!)                           --
------------------------------------------------------------------------------------
--~ setmetatable(_ENV,{
  --~ __newindex=function (self,key,value) --locked_global_write
    --~ error('\n\n[ER Global Lock] Forbidden global *write*:\n'
      --~ .. serpent.line{key=key or '<nil>',value=value or '<nil>'}..'\n')
    --~ end,
  --~ __index   =function (self,key) --locked_global_read
    --~ error('\n\n[ER Global Lock] Forbidden global *read*:\n'
      --~ .. serpent.line{key=key or '<nil>'}..'\n')
    --~ end ,
  --~ })
