------------------------------------------------------------------------------------
-- "AAI Programmable Vehicles", "Vehicle Wagons 2" and possibly other mods replace
-- vehicle entities, so we need to keep our lists updated. Until now, we just stored
-- the vehicle entities, without keeping track of the unit_number. That was enough
-- as long as entities were supposed to be valid, but it's hard or even impossible
-- to identify replaced entities.
-- This script rebuilds the lists, using the unit number as index of vehicle data.

local GCKI = require('__GCKI__.common')('GCKI')
local script_name = "00-18-03_combine_locked_+_claimed_vehicle_lists.lua"

GCKI.dprint("Entered migration script " .. script_name .. ".")

-- Only act if the original lists still exist. These will be removed after migration,
-- once they've been consolidated into a new, restructured list (global.GCKI_vehicles).
if global.vehicles and global.locked_vehicles and not global.GCKI_vehicles then

    -- Get the entity of the player who locked this vehicle
    local function is_locked_vehicle(vehicle)
        --~ GCKI.dprint("Entered function is_locked_vehicle(" .. tostring(vehicle.name) .. ")")
        local locker = nil

        -- Unlocked vehicle
        if (vehicle.operable and vehicle.active and vehicle.minable) then
            GCKI.dprint({"GCKI-messages.lock_state", vehicle.name, {"GCKI-states.unlocked"}})
        -- Locked vehicle
        else
            --~ GCKI.dprint({"GCKI-messages.lock_state", GCKI.loc_name(vehicle), {"GCKI-states.locked"}})
            -- Find the player who locked it
            for player_index, vehicles in pairs(global.locked_vehicles) do
                for _, vehicle2 in pairs(vehicles) do
                    if vehicle == vehicle2 then
                        locker = game.players[player_index]
                        GCKI.dprint(vehicle.name .. " has been locked by " .. locker.name ..
                                    " (player " .. locker.index .. ")!")
                        break
                    end
                end
                if locker then break end
            end
        end
        --~ GCKI.dprint("End of function is_locked_vehicle(" .. tostring(vehicle.name) .. ")")
        return locker and locker.index
    end


    -- Initialize new list
    global.GCKI_vehicles = {}

    -- Transfer list of owned vehicles to new list
    for player, vehicle in pairs(global.vehicles) do
        if vehicle.valid then
            GCKI.dprint("Adding " .. vehicle.name .. " (" .. vehicle.unit_number .. ") to list.")
            global.GCKI_vehicles[vehicle.unit_number] = {
                    ["entity"] = vehicle,
                    ["owner"] = player,
                    ["locker"] = is_locked_vehicle(vehicle),
            }
        end
    end
    GCKI.dprint("Added owned vehicles to new list \"global.GCKI_vehicles\".")

    -- Transfer list of locked vehicles to new list
    for player, table in pairs(global.locked_vehicles) do
        for _, vehicle in pairs(table) do
            -- Vehicle may already be in list if it also has been claimed by somebody!
            if vehicle.valid and not global.GCKI_vehicles[vehicle.unit_number] then
                GCKI.dprint({"", "Adding ", GCKI.loc_name(vehicle), " (", vehicle.unit_number, ") to list!"})
                global.GCKI_vehicles[vehicle.unit_number] = {
                    ["entity"] = vehicle,
                    ["owner"] = nil,
                    ["locker"] = player,
                }
            end
        end
    end
    GCKI.dprint("Added locked vehicles to new list \"global.GCKI_vehicles\".")

    -- Remove original lists
    global.locked_vehicles = nil
    global.vehicles = nil
    GCKI.dprint("Removed old lists \"global.locked_vehicles\" and \"global.GCKI_vehicles\". " .. serpent.block(global))
end

GCKI.dprint("End of migration script " .. script_name .. ".")
